from __future__ import annotations

import weakref
from ctypes import (
    POINTER,
    byref,
    c_buffer,
    c_char,
    c_char_p,
    c_int,
    cast,
    create_string_buffer,
    pointer,
)
from typing import TYPE_CHECKING

from pyglet import gl

from .uniform import Uniform, UniformBlock

if TYPE_CHECKING:
    from arcade.gl import Context


class ComputeShader:
    """
    A higher level wrapper for an OpenGL compute shader.

    Args:
        ctx:
            The context this shader belongs to.
        glsl_source:
            The GLSL source code for the compute shader.
    """

    def __init__(self, ctx: Context, glsl_source: str) -> None:
        self._ctx = ctx
        self._source = glsl_source
        self._uniforms: dict[str, UniformBlock | Uniform] = dict()

        from arcade.gl import ShaderException

        # Create the program
        self._glo = glo = gl.glCreateProgram()
        if not self._glo:
            raise ShaderException("Failed to create program object")

        self._shader_obj = gl.glCreateShader(gl.GL_COMPUTE_SHADER)
        if not self._shader_obj:
            raise ShaderException("Failed to create compute shader object")

        # Set source
        source_bytes = self._source.encode("utf-8")
        strings = byref(cast(c_char_p(source_bytes), POINTER(c_char)))
        lengths = pointer(c_int(len(source_bytes)))
        gl.glShaderSource(self._shader_obj, 1, strings, lengths)

        # Compile and check result
        gl.glCompileShader(self._shader_obj)
        result = c_int()
        gl.glGetShaderiv(self._shader_obj, gl.GL_COMPILE_STATUS, byref(result))
        if result.value == gl.GL_FALSE:
            msg = create_string_buffer(512)
            length = c_int()
            gl.glGetShaderInfoLog(self._shader_obj, 512, byref(length), msg)
            raise ShaderException(
                (
                    f"Error compiling compute shader "
                    f"({result.value}): {msg.value.decode('utf-8')}\n"
                    f"---- [compute shader] ---\n"
                )
                + "\n".join(
                    f"{str(i + 1).zfill(3)}: {line} "
                    for i, line in enumerate(self._source.split("\n"))
                )
            )

        # Attach and link shader
        gl.glAttachShader(self._glo, self._shader_obj)
        gl.glLinkProgram(self._glo)
        gl.glDeleteShader(self._shader_obj)
        status = c_int()
        gl.glGetProgramiv(self._glo, gl.GL_LINK_STATUS, status)
        if not status.value:
            length = c_int()
            gl.glGetProgramiv(self._glo, gl.GL_INFO_LOG_LENGTH, length)
            log = c_buffer(length.value)
            gl.glGetProgramInfoLog(self._glo, len(log), None, log)
            raise ShaderException("Program link error: {}".format(log.value.decode()))

        self._introspect_uniforms()
        self._introspect_uniform_blocks()

        if self._ctx.gc_mode == "auto":
            weakref.finalize(self, ComputeShader.delete_glo, self._ctx, glo)

        ctx.stats.incr("compute_shader")

    @property
    def glo(self) -> int:
        """The name/id of the OpenGL resource"""
        return self._glo

    def _use(self) -> None:
        """
        Use/activate the compute shader.

        .. Note::

            This is not necessary to call in normal use cases
            since ``run()`` already does this for you.
        """
        gl.glUseProgram(self._glo)
        self._ctx.active_program = self

    def run(self, group_x=1, group_y=1, group_z=1) -> None:
        """
        Run the compute shader.

        When running a compute shader we specify how many work groups should
        be executed on the ``x``, ``y`` and ``z`` dimension. The size of the work group
        is defined in the compute shader.

        .. code:: glsl

            // Work group with one dimension. 16 work groups executed.
            layout(local_size_x=16) in;
            // Work group with two dimensions. 256 work groups executed.
            layout(local_size_x=16, local_size_y=16) in;
            // Work group with three dimensions. 4096 work groups executed.
            layout(local_size_x=16, local_size_y=16, local_size_z=16) in;

        Group sizes are ``1`` by default. If your compute shader doesn't specify
        a size for a dimension or uses ``1`` as size you don't have to supply
        this parameter.

        Args:
            group_x: The number of work groups to be launched in the X dimension.
            group_y: The number of work groups to be launched in the y dimension.
            group_z: The number of work groups to be launched in the z dimension.
        """
        self._use()
        gl.glDispatchCompute(group_x, group_y, group_z)

    def __getitem__(self, item) -> Uniform | UniformBlock:
        """Get a uniform or uniform block"""
        try:
            uniform = self._uniforms[item]
        except KeyError:
            raise KeyError(f"Uniform with the name `{item}` was not found.")

        return uniform.getter()

    def __setitem__(self, key, value):
        """Set a uniform value"""
        # Ensure we are setting the uniform on this program
        # if self._ctx.active_program != self:
        #     self.use()

        try:
            uniform = self._uniforms[key]
        except KeyError:
            raise KeyError(f"Uniform with the name `{key}` was not found.")

        uniform.setter(value)

    def __hash__(self) -> int:
        return id(self)

    def __del__(self):
        if self._ctx.gc_mode == "context_gc" and self._glo > 0:
            self._ctx.objects.append(self)

    def delete(self) -> None:
        """
        Destroy the internal compute shader object.

        This is normally not necessary, but depends on the
        garbage collection configured in the context.
        """
        ComputeShader.delete_glo(self._ctx, self._glo)
        self._glo = 0

    @staticmethod
    def delete_glo(ctx, prog_id):
        """
        Low level method for destroying a compute shader by id.

        Args:
            ctx: The context this program belongs to.
            prog_id: The OpenGL id of the program.
        """
        # Check to see if the context was already cleaned up from program
        # shut down. If so, we don't need to delete the shaders.
        if gl.current_context is None:
            return

        gl.glDeleteProgram(prog_id)
        # TODO: Count compute shaders
        ctx.stats.decr("compute_shader")

    def _introspect_uniforms(self):
        """Figure out what uniforms are available and build an internal map"""
        # Number of active uniforms in the program
        active_uniforms = gl.GLint(0)
        gl.glGetProgramiv(self._glo, gl.GL_ACTIVE_UNIFORMS, byref(active_uniforms))

        # Loop all the active uniforms
        for index in range(active_uniforms.value):
            # Query uniform information like name, type, size etc.
            u_name, u_type, u_size = self._query_uniform(index)
            u_location = gl.glGetUniformLocation(self._glo, u_name.encode())

            # Skip uniforms that may be in Uniform Blocks
            # TODO: We should handle all uniforms
            if u_location == -1:
                # print(f"Uniform {u_location} {u_name} {u_size} {u_type} skipped")
                continue

            u_name = u_name.replace("[0]", "")  # Remove array suffix
            self._uniforms[u_name] = Uniform(
                self._ctx, self._glo, u_location, u_name, u_type, u_size
            )

    def _introspect_uniform_blocks(self):
        """Finds uniform blocks and maps the to python objectss"""
        active_uniform_blocks = gl.GLint(0)
        gl.glGetProgramiv(self._glo, gl.GL_ACTIVE_UNIFORM_BLOCKS, byref(active_uniform_blocks))
        # print('GL_ACTIVE_UNIFORM_BLOCKS', active_uniform_blocks)

        for loc in range(active_uniform_blocks.value):
            index, size, name = self._query_uniform_block(loc)
            block = UniformBlock(self._glo, index, size, name)
            self._uniforms[name] = block

    def _query_uniform(self, location: int) -> tuple[str, int, int]:
        """Retrieve Uniform information at given location.

        Returns the name, the type as a GLenum (GL_FLOAT, ...) and the size. Size is
        greater than 1 only for Uniform arrays, like an array of floats or an array
        of Matrices.
        """
        u_size = gl.GLint()
        u_type = gl.GLenum()
        buf_size = 192  # max uniform character length
        u_name = create_string_buffer(buf_size)
        gl.glGetActiveUniform(
            self._glo,  # program to query
            location,  # location to query
            buf_size,  # size of the character/name buffer
            None,  # the number of characters actually written by OpenGL in the string
            u_size,  # size of the uniform variable
            u_type,  # data type of the uniform variable
            u_name,  # string buffer for storing the name
        )
        return u_name.value.decode(), u_type.value, u_size.value

    def _query_uniform_block(self, location: int) -> tuple[int, int, str]:
        """Query active uniform block by retrieving the name and index and size"""
        # Query name
        u_size = gl.GLint()
        buf_size = 192  # max uniform character length
        u_name = create_string_buffer(buf_size)
        gl.glGetActiveUniformBlockName(
            self._glo,  # program to query
            location,  # location to query
            256,  # max size if the name
            u_size,  # length
            u_name,
        )
        # Query index
        index = gl.glGetUniformBlockIndex(self._glo, u_name)
        # Query size
        b_size = gl.GLint()
        gl.glGetActiveUniformBlockiv(self._glo, index, gl.GL_UNIFORM_BLOCK_DATA_SIZE, b_size)
        return index, b_size.value, u_name.value.decode()



from __future__ import annotations

import math
from collections.abc import Generator, Iterable
from typing import TYPE_CHECKING

from pyglet.event import EVENT_HANDLED, EVENT_UNHANDLED

from arcade import get_window
from arcade.camera.default import DefaultProjector
from arcade.types.rect import LRBT, Rect

if TYPE_CHECKING:
    from arcade import View
    from arcade.camera import Projector

__all__ = ["Section", "SectionManager"]


class Section:
    """
    A Section represents a rectangular portion of the viewport
    Events are dispatched to the section based on it's position on the screen.

    A section can only be added to a single SectionManager.

    Args:
        left: the left position of this section
        bottom: the bottom position of this section
        width: the width of this section
        height: the height of this section
        name: the name of this section
        accept_keyboard_keys:
            bool | Iterable accept_keyboard_keys: whether this section
                captures keyboard keys through. keyboard events. If the param is an iterable
                means the keyboard keys that are captured in press/release events:
                for example: ``[arcade.key.UP, arcade.key.DOWN]`` will only capture this two keys
        accept_mouse_events:
            bool  Iterable accept_mouse_events: whether this section
                captures mouse events. If the param is an iterable means the mouse events
                that are captured. for example: ``['on_mouse_press', 'on_mouse_release']``
                will only capture this two events.
        prevent_dispatch: a list of event names that will not be dispatched to subsequent
            sections. You can pass None (default) or {True} to prevent the dispatch of all events.
        prevent_dispatch_view: a list of event names that will not be dispatched to the view.
            You can pass None (default) or {True} to prevent the dispatch of all events to the view.
        local_mouse_coordinates: if True the section mouse events will receive x, y
            coordinates section related to the section dimensions and position (not related
            to the screen)
        enabled: if False the section will not capture any events
        modal: if True the section will be a modal section: will prevent updates
            and event captures on other sections. Will also draw last (on top) but capture
            events first.
        draw_order: The order this section will have when on_draw is called.
            The lower the number the earlier this will get draw.
            This can be different from the event capture order or the on_update order which
            is defined by the insertion order.
    """

    def __init__(
        self,
        left: int | float,
        bottom: int | float,
        width: int | float,
        height: int | float,
        *,
        name: str | None = None,
        accept_keyboard_keys: bool | Iterable = True,
        accept_mouse_events: bool | Iterable = True,
        prevent_dispatch: Iterable | bool | None = None,
        prevent_dispatch_view: Iterable | bool | None = None,
        local_mouse_coordinates: bool = False,
        enabled: bool = True,
        modal: bool = False,
        draw_order: int = 1,
    ):
        self.name: str | None = name
        """The name of this section"""

        # parent view: set by the SectionManager. Protected, you should not change
        # section.view manually
        self._view: View | None = None
        self._section_manager: SectionManager | None = None

        # section options
        self._enabled: bool = enabled  # enables or disables this section
        # prevent the following sections from receiving input events and updating
        self._modal: bool = modal

        # set draw_order: the lower the number the earlier this section will get draw
        self._draw_order: int = draw_order

        self.block_updates: bool = False
        """if True 'update' and 'on_update' will not trigger in this section"""

        self.accept_keyboard_keys: bool | Iterable = accept_keyboard_keys
        """Arcade keyboard keys to accept."""

        self.accept_mouse_events: bool | Iterable = accept_mouse_events
        """Arcade mouse events to accept."""

        if isinstance(prevent_dispatch, bool):
            prevent_dispatch = {prevent_dispatch}
        self.prevent_dispatch: Iterable = {True} if prevent_dispatch is None else prevent_dispatch
        assert isinstance(self.prevent_dispatch, Iterable)
        """prevents events to propagate"""

        if isinstance(prevent_dispatch_view, bool):
            prevent_dispatch_view = {prevent_dispatch_view}
        self.prevent_dispatch_view: Iterable = (
            {True} if prevent_dispatch_view is None else prevent_dispatch_view
        )
        assert isinstance(self.prevent_dispatch_view, Iterable)
        """prevents events to propagate to the view"""

        self.local_mouse_coordinates: bool = local_mouse_coordinates
        """mouse coordinates relative to section"""

        # section position into the current viewport
        # if screen is resized it's upto the user to move or resize each section
        # (section will receive on_resize event)
        self._left: int | float = left
        self._bottom: int | float = bottom
        self._width: int | float = width
        self._height: int | float = height
        self._right: int | float = left + width
        self._top: int | float = bottom + height

        # section event capture dimensions
        # if section is modal, capture all events on the screen
        self._ec_left: int | float = 0 if self._modal else self._left
        self._ec_right: int | float = self.window.width if self._modal else self._right
        self._ec_bottom: int | float = 0 if self._modal else self._bottom
        self._ec_top: int | float = self.window.height if self._modal else self._top

        self.camera: Projector | None = None
        """optional section camera"""

    def __repr__(self):
        name = f"Section {self.name}" if self.name else "Section"
        dimensions = (self.left, self.right, self.top, self.bottom)
        return f"{name} at {dimensions} "

    @property
    def view(self):
        """The view this section is set on"""
        return self._view

    @property
    def section_manager(self) -> SectionManager | None:
        """Returns the section manager this section is added to"""
        return self._section_manager

    @property
    def enabled(self) -> bool:
        """Enables or disables this section"""
        return self._enabled

    @enabled.setter
    def enabled(self, value: bool):
        if value is self._enabled:
            return
        self._enabled = value
        if value:
            self.on_show_section()
        else:
            self.on_hide_section()

    @property
    def modal(self) -> bool:
        """
        Returns the modal state (Prevent the following sections from
        receiving input events and updating)
        """
        return self._modal

    @property
    def draw_order(self) -> int:
        """
        Returns the draw order state
        The lower the number the earlier this section will get draw
        """
        return self._draw_order

    @property
    def left(self) -> int | float:
        """Left edge of this section"""
        return self._left

    @left.setter
    def left(self, value: int):
        self._left = value
        self._right = value + self._width
        self._ec_left = 0 if self._modal else value
        self._ec_right = self.window.width if self._modal else self._right

    @property
    def bottom(self) -> int | float:
        """The bottom edge of this section"""
        return self._bottom

    @bottom.setter
    def bottom(self, value: int):
        self._bottom = value
        self._top = value + self._height
        self._ec_bottom = 0 if self._modal else value
        self._ec_top = self.window.height if self._modal else self._top

    @property
    def width(self) -> int | float:
        """The width of this section"""
        return self._width

    @width.setter
    def width(self, value: int):
        self._width = value
        self._right = value + self._left
        self._ec_right = self.window.width if self._modal else self._right

    @property
    def height(self) -> int | float:
        """The height of this section"""
        return self._height

    @height.setter
    def height(self, value: int):
        self._height = value
        self._top = value + self._bottom
        self._ec_top = self.window.height if self._modal else self._top

    @property
    def right(self) -> int | float:
        """Right edge of this section"""
        return self._right

    @right.setter
    def right(self, value: int):
        self._right = value
        self._left = value - self._width
        self._ec_right = self.window.width if self._modal else value
        self._ec_left = 0 if self._modal else self._left

    @property
    def top(self) -> int | float:
        """Top edge of this section"""
        return self._top

    @top.setter
    def top(self, value: int):
        self._top = value
        self._bottom = value - self._height
        self._ec_top = self.window.height if self._modal else value
        self._ec_bottom = 0 if self._modal else self._bottom

    @property
    def rect(self) -> Rect:
        """The section rectangle of this view"""
        return LRBT(self.left, self.right, self.bottom, self.top)

    @property
    def window(self):
        """The view window"""
        if getattr(self, "_view", None) is None or self._view is None:
            return get_window()
        else:
            return self._view.window

    def overlaps_with(self, section: Section) -> bool:
        """Checks if this section overlaps with another section

        Args:
            section: The section to check for overlap
        """
        return not (
            self.right < section.left
            or self.left > section.right
            or self.top < section.bottom
            or self.bottom > section.top
        )

    def mouse_is_on_top(self, x: int, y: int) -> bool:
        """Check if the current mouse position is on top of this section

        Args:
            x: The x position of the mouse
            y: The y position of the mouse
        """
        test_x = self._left <= x <= self._right
        test_y = self._bottom <= y <= self._top
        return test_x and test_y

    def should_receive_mouse_event(self, x: int, y: int) -> bool:
        """Check if the current section should receive a mouse event at a given position

        Args:
            x: The x position of the mouse
            y: The y position of the mouse
        """
        test_x = self._ec_left <= x <= self._ec_right
        test_y = self._ec_bottom <= y <= self._ec_top
        return test_x and test_y

    def get_xy_screen_relative(self, section_x: int, section_y: int):
        """Returns screen coordinates from section coordinates

        Args:
            section_x: The x position of the section
            section_y: The y position of the section
        """
        return self.left + section_x, self.bottom + section_y

    def get_xy_section_relative(self, screen_x: int, screen_y: int):
        """returns section coordinates from screen coordinates

        Args:
            screen_x: The x position of the screen
            screen_y: The y position of the screen
        """
        return screen_x - self.left, screen_y - self.bottom

    # Following methods are just the usual view methods
    #  + on_mouse_enter / on_mouse_leave / on_show_section / on_hide_section

    def on_draw(self):
        """Override this method with your custom drawing code."""
        pass

    def on_update(self, delta_time: float):
        """Override this method with your custom update code."""
        pass

    def on_resize(self, width: int, height: int):
        """Override this method with your custom resize code."""
        pass

    def on_mouse_press(self, x: int, y: int, button: int, modifiers: int):
        """
        Called when the user presses a mouse button.

        Args:
            x: x position of the mouse
            y: y position of the mouse
            button: the button pressed
            modifiers: the modifiers pressed
        """
        pass

    def on_mouse_release(self, x: int, y: int, button: int, modifiers: int):
        """
        Called when the user releases a mouse button.

        Args:
            x: x position of the mouse
            y: y position of the mouse
            button: the button released
            modifiers: the modifiers pressed
        """
        pass

    def on_mouse_motion(self, x: int, y: int, dx: int, dy: int):
        """
        Called when the user moves the mouse.

        Args:
            x: x position of the mouse
            y: y position of the mouse
            dx: change in x position
            dy: change in y position
        """
        pass

    def on_mouse_scroll(self, x: int, y: int, scroll_x: int, scroll_y: int):
        """
        Called when the user scrolls the mouse wheel.

        Args:
            x: x position of the mouse
            y: y position of the mouse
            scroll_x: change in x position
            scroll_y: change in y position
        """
        pass

    def on_mouse_drag(self, x: int, y: int, dx: int, dy: int, _buttons: int, _modifiers: int):
        """
        Called when the user moves the mouse with a button pressed.

        Args:
            x: x position of the mouse
            y: y position of the mouse
            dx: change in x position
            dy: change in y position
        """
        self.on_mouse_motion(x, y, dx, dy)

    def on_mouse_enter(self, x: int, y: int):
        """
        Called when the mouse enters the section

        Args:
            x: x position of the mouse
            y: y position of the mouse
        """
        pass

    def on_mouse_leave(self, x: int, y: int):
        """
        Called when the mouse leaves the section

        Args:
            x: x position of the mouse
            y: y position of the mouse
        """
        pass

    def on_key_press(self, symbol: int, modifiers: int):
        """
        Called when the user presses a key.

        Args:
            symbol: the key pressed
            modifiers: the modifiers pressed
        """
        pass

    def on_key_release(self, symbol: int, modifiers: int):
        """
        Called when the user releases a key.

        Args:
            symbol: the key released
            modifiers: the modifiers pressed
        """
        pass

    def on_show_section(self):
        """Called when the section is shown"""
        pass

    def on_hide_section(self):
        """Called when the section is hidden"""
        pass


class SectionManager:
    """
    This manages the different Sections a View has.
    Actions such as dispatching the events to the correct Section, draw order, etc.

    Args:
        view: the view this section manager belongs to
    """

    def __init__(self, view: View):
        self.view = view  # the view this section manager belongs to

        # store sections in update/event order and in draw order
        # a list of the current sections for this in update/event order
        self._sections: list[Section] = []

        # the list of current sections in draw order
        self._sections_draw: list[Section] = []

        # generic camera to reset after a custom camera is use
        # this camera is set to the whole viewport
        self.camera: DefaultProjector = DefaultProjector()

        # Holds the section the mouse is currently on top
        self.mouse_over_sections: list[Section] = []

        # True will call view.on_draw before sections on_draw, False after,
        # None will not call view on_draw
        self.view_draw_first: bool | None = True
        # True will call view.on_update before sections on_update, False after,
        # None will not call view on_update
        self.view_update_first: bool | None = True
        # True will call view.on_resize before sections on_resize, False after,
        # None will not call view on_resize
        self.view_resize_first: bool | None = True

        # Events that the section manager should handle (instead of the View) if
        # sections are present in a View
        self.managed_events: set = {
            "on_mouse_motion",
            "on_mouse_drag",
            "on_mouse_press",
            "on_mouse_release",
            "on_mouse_scroll",
            "on_mouse_enter",
            "on_mouse_leave",
            "on_key_press",
            "on_key_release",
            "on_draw",
            "on_update",
            "on_resize",
        }

    @property
    def sections(self) -> list[Section]:
        """Property that returns the list of sections"""
        return self._sections

    @property
    def has_sections(self) -> bool:
        """Returns true if this section manager has sections"""
        return bool(self._sections)

    @property
    def is_current_view(self) -> bool:
        """
        Returns if this section manager view is the current on the view window
        a.k.a.: is the view that is currently being shown
        """
        return self.view.window.current_view is self.view

    def disable(self) -> None:
        """
        Disable all sections
        Disabling a section will trigger section.on_hide_section
        """
        self.view.window.remove_handlers(self)

        for section in self._sections:
            section.enabled = False

    def enable(self) -> None:
        """
        Registers event handlers to the window and enables all sections.

        Enabling a section will trigger section.on_show_section
        """
        section_handlers = {
            event_type: getattr(self, event_type, None) for event_type in self.managed_events
        }
        if section_handlers:
            self.view.window.push_handlers(**section_handlers)

        for section in self.sections:
            # sections are enabled by default
            if section.enabled:
                section.on_show_section()
            else:
                section.enabled = True
                # enabling a section will trigger on_show_section

    def get_section_by_name(self, name: str) -> Section | None:
        """
        Returns the first section with the given name

        Args:
            name: The name of the section you want
        Returns:
            The first section with the provided name. None otherwise
        """
        for section in self._sections:
            if section.name == name:
                return section
        return None

    def add_section(
        self,
        section: Section,
        at_index: int | None = None,
        at_draw_order: int | None = None,
    ) -> None:
        """
        Adds a section to this Section Manager
        Will trigger section.on_show_section if section is enabled

        Args:
            section: The section to add to this section manager
            at_index: Inserts the section at that index for event capture and update events.
                If None at the end
            at_draw_order: Inserts the section in a specific draw order.
                Overwrites section.draw_order
        """
        if not isinstance(section, Section):
            raise ValueError("You can only add Section instances")
        if section.section_manager is not None:
            raise ValueError("Section is already added to a SectionManager")

        # set the view and section manager
        section._view = self.view
        section._section_manager = self

        if at_index is None:
            self._sections.append(section)
        else:
            self._sections.insert(at_index, section)
        # keep sections order updated in the lists of sections to draw
        self.sort_section_event_order()
        if at_draw_order is None:
            self.sort_sections_draw_order()
        else:
            section._draw_order = at_draw_order  # this will trigger self.sort_section_draw_order
            self.sort_sections_draw_order()

        # trigger on_show_section if the view is the current one and section is enabled:
        if self.is_current_view and section.enabled:
            section.on_show_section()

    def remove_section(self, section: Section) -> None:
        """
        Removes a section from this section manager

        Args:
            section: The section to remove
        """

        # trigger on_hide_section if the view is the current one and section is enabled
        if self.is_current_view and section.enabled:
            section.on_hide_section()

        section._view = None
        self._sections.remove(section)

        # keep sections order updated in the lists of sections
        self.sort_section_event_order()
        self.sort_sections_draw_order()

    def sort_section_event_order(self) -> None:
        """
        This will sort sections on event capture order (and update) based on
        insertion order and section.modal.
        """
        # modals go first
        self._sections.sort(key=lambda s: 0 if s.modal else 1)

    def sort_sections_draw_order(self) -> None:
        """This will sort sections on draw order based on section.draw_order and section.modal"""
        # modals go last
        self._sections_draw = sorted(
            self._sections, key=lambda s: math.inf if s.modal else s.draw_order
        )

    def clear_sections(self) -> None:
        """Removes all sections and calls on_hide_section for each one if enabled"""
        for section in self._sections:
            if section.enabled:
                section.on_hide_section()
            section._view = None
        self._sections = []
        self._sections_draw = []

    def on_update(self, delta_time: float) -> bool:
        """
        Called on each event loop.

        Args:
            delta_time: the delta time since this method was called last time
        """
        modal_present = False
        if self.view_update_first is True:
            self.view.on_update(delta_time)
        for section in self._sections:
            if section.enabled and not section.block_updates and not modal_present:
                section.on_update(delta_time)
                if section.modal:
                    modal_present = True
        if self.view_update_first is False:
            self.view.on_update(delta_time)

        # prevent further event dispatch to the view, which was already invoked
        return True

    def on_draw(self) -> bool:
        """
        Called on each event loop to draw.

        It automatically calls camera.use() for each section that has a camera and
        resets the camera effects by calling the default SectionManager camera
        afterward if needed. The SectionManager camera defaults to a camera that
        has the viewport and projection for the whole screen.

        This method will consume the on_draw event, to prevent further on_draw calls on the view.
        """
        if self.view_draw_first is True:
            self.view.on_draw()
        for section in self._sections_draw:  # iterate over sections_draw
            if not section.enabled:
                continue
            if section.camera:
                # use the camera of the current section before section.on_draw
                section.camera.use()
            section.on_draw()
            if section.camera:
                # reset to the default camera after the section is drawn
                self.camera.use()
        if self.view_draw_first is False:
            self.view.on_draw()

        # prevent further event dispatch to the view, which was already invoked
        return True

    def on_resize(self, width: int, height: int) -> bool:
        """
        Called when the window is resized.

        Args:
            width: the new width of the screen
            height: the new height of the screen
        """
        # The Default camera auto-resizes.
        if self.view_resize_first is True:
            self.view.on_resize(width, height)  # call resize on the view
        for section in self.sections:
            if section.enabled:
                section.on_resize(width, height)
        if self.view_resize_first is False:
            self.view.on_resize(width, height)  # call resize on the view

        # prevent further event dispatch to the view, which was already invoked
        return True

    def disable_all_keyboard_events(self) -> None:
        """Removes the keyboard event handling from all sections"""
        for section in self.sections:
            section.accept_keyboard_keys = False

    def get_first_section(self, x: int, y: int, *, event_capture: bool = True) -> Section | None:
        """
        Returns the first section based on x,y position

        Args:
            x: the x axis coordinate
            y: the y axis coordinate
            event_capture: True will use event capture dimensions,
                False will use section draw size
        Returns:
            A section if match the params otherwise None
        """
        for section in self._sections:
            if section.enabled:
                if event_capture is True and section.should_receive_mouse_event(x, y):
                    return section
                if event_capture is False and section.mouse_is_on_top(x, y):
                    return section
        return None

    def get_sections(
        self, x: int, y: int, *, event_capture: bool = True
    ) -> Generator[Section, None, None]:
        """
        Returns a list of sections based on x,y position

        Args:
            x: the x axis coordinate
            y: the y axis coordinate
            event_capture: True will use event capture dimensions,
                False will use section draw size
        Returns:
            A generator with the sections that match the params
        """
        for section in self._sections:
            if section.enabled:
                if event_capture is True and section.should_receive_mouse_event(x, y):
                    yield section
                if event_capture is False and section.mouse_is_on_top(x, y):
                    yield section

    def dispatch_mouse_event(
        self,
        event: str,
        x: int,
        y: int,
        *args,
        current_section: Section | None = None,
        **kwargs,
    ) -> bool | None:
        """
        Generic method to dispatch mouse events to the correct Sections

        Args:
            event: the mouse event name to dispatch
            x: the x axis coordinate
            y: the y axis coordinate
            args: any other position arguments that should be delivered to the dispatched event
            current_section: the section this mouse event should be delivered to.
                If None, will retrieve all
                sections that should receive this event based on x, y coordinates
            kwargs: any other keyword arguments that should be delivered to the dispatched event
        Returns:
            ``EVENT_HANDLED`` or ``EVENT_UNHANDLED``, or whatever the dispatched method returns
        """

        sections: list | Generator

        if current_section:
            # affected section is already pre-computed
            sections = [current_section]
        else:
            # get the sections from mouse position
            sections = self.get_sections(x, y)

        prevent_dispatch = EVENT_UNHANDLED
        prevent_dispatch_view = EVENT_UNHANDLED

        for section in sections:
            if prevent_dispatch is EVENT_HANDLED:
                break
            mouse_events_allowed = section.accept_mouse_events
            if mouse_events_allowed is False:
                continue
            if mouse_events_allowed is True or event in mouse_events_allowed:  # event allowed
                # get the method to call from the section
                method = getattr(section, event, None)
                if method:
                    if section.local_mouse_coordinates:
                        position = section.get_xy_section_relative(x, y)
                    else:
                        position = x, y

                    # call the section method
                    prevent_dispatch = method(*position, *args, **kwargs)

                    # mark prevent dispatch as handled if section is modal
                    prevent_dispatch = EVENT_HANDLED if section.modal else prevent_dispatch

                    # check if section prevents dispatching this event any further
                    # in the section stack
                    if prevent_dispatch is EVENT_HANDLED or any(
                        test in section.prevent_dispatch for test in [True, event]
                    ):
                        # prevent_dispatch attributte from section only affects if
                        #  the method is implemented in the same section
                        prevent_dispatch = EVENT_HANDLED

            if any(test in section.prevent_dispatch_view for test in [True, event]):
                # check if the section prevents dispatching events to the view
                prevent_dispatch_view = EVENT_HANDLED

        if prevent_dispatch_view is EVENT_UNHANDLED:
            # call the method from the view.
            method = getattr(self.view, event, None)  # get the method from the view
            if method:
                # call the view method
                prevent_dispatch_view = method(x, y, *args, **kwargs)

        return prevent_dispatch_view or prevent_dispatch

    def dispatch_keyboard_event(self, event: str, *args, **kwargs) -> bool | None:
        """
        Generic method to dispatch keyboard events to the correct sections

        Args:
            event: the keyboard event name to dispatch
            args: any other position arguments that should be delivered to the dispatched event
            kwargs: any other keyword arguments that should be delivered to the dispatched event
        Returns:
            ``EVENT_HANDLED`` or ``EVENT_UNHANDLED``, or whatever the dispatched method returns
        """
        propagate_to_view = True
        prevent_dispatch = EVENT_UNHANDLED
        for section in self.sections:
            if prevent_dispatch:
                break
            if not section.enabled:
                continue
            keys_allowed = section.accept_keyboard_keys
            if keys_allowed is False:
                continue
            if keys_allowed is True or args[0] in keys_allowed or args in keys_allowed:
                if any(test in section.prevent_dispatch_view for test in [True, event]):
                    propagate_to_view = False
                # get the method to call from the section
                method = getattr(section, event, None)
                if method:
                    # call the section method
                    prevent_dispatch = method(*args, **kwargs)
                    if prevent_dispatch is EVENT_HANDLED or any(
                        test in section.prevent_dispatch for test in [True, event]
                    ):
                        # prevent_dispatch attributes from section only affect
                        #  if the method is implemented in the same section
                        prevent_dispatch = EVENT_HANDLED
            if section.modal:
                # if this section is modal, then avoid passing any event
                # to more sections
                return prevent_dispatch

        if propagate_to_view is False:
            return prevent_dispatch

        method = getattr(self.view, event, None)  # get the method from the view
        if method:
            # call the view method
            return method(*args, **kwargs) or prevent_dispatch
        return EVENT_UNHANDLED

    def on_mouse_press(self, x: int, y: int, *args, **kwargs) -> bool | None:
        """
        Triggers the on_mouse_press event on the appropriate sections or view

        Args:
            x: the x axis coordinate
            y: the y axis coordinate
            args: any other position arguments that should be delivered to the dispatched event
            kwargs: any other keyword arguments that should be delivered to the dispatched event
        Returns:
            ``EVENT_HANDLED`` or ``EVENT_UNHANDLED``, or whatever the dispatched method returns
        """
        return self.dispatch_mouse_event("on_mouse_press", x, y, *args, **kwargs)

    def on_mouse_release(self, x: int, y: int, *args, **kwargs) -> bool | None:
        """
        Triggers the on_mouse_release event on the appropriate sections or view

        Args:
            x: the x axis coordinate
            y: the y axis coordinate
            args: any other position arguments that should be delivered to the dispatched event
            kwargs: any other keyword arguments that should be delivered to the dispatched event
        Returns:
            ``EVENT_HANDLED`` or ``EVENT_UNHANDLED``, or whatever the dispatched method returns
        """
        return self.dispatch_mouse_event("on_mouse_release", x, y, *args, **kwargs)

    def dispatch_mouse_enter_leave_events(
        self, event_origin: str, x: int, y: int, *args, **kwargs
    ) -> bool | None:
        """
        This helper method will dispatch mouse enter / leave events to sections
        based on 'on_mouse_motion' and 'on_mouse_drag' events.
        Will also dispatch the event (event_origin) that called this method

        Args:
            event_origin: The mouse event name that called this method.
                This event will be called here.
            x: The x axis coordinate
            y: The y axis coordinate
            args: Any other position arguments that should be delivered to the dispatched event
            kwargs: Any other keyword arguments that should be delivered to the dispatched event
        Returns:
            ``EVENT_HANDLED`` or ``EVENT_UNHANDLED``, or whatever the dispatched method returns
        """
        before_sections = self.mouse_over_sections
        current_sections = list(self.get_sections(x, y))  # consume the generator

        prevent_dispatch_origin = EVENT_UNHANDLED  # prevent dispatch for the origin mouse event

        prevent_dispatch_el = EVENT_UNHANDLED  # prevent dispatch for enter/leave events
        for section in before_sections:
            if section not in current_sections:
                if prevent_dispatch_el is EVENT_HANDLED:
                    break
                # dispatch on_mouse_leave to before_section
                prevent_dispatch_el = self.dispatch_mouse_event(
                    "on_mouse_leave", x, y, current_section=section
                )

        prevent_dispatch_el = EVENT_UNHANDLED
        for section in current_sections:
            if section not in before_sections:
                if prevent_dispatch_el is EVENT_UNHANDLED:
                    # dispatch on_mouse_enter to current_section
                    prevent_dispatch_el = self.dispatch_mouse_event(
                        "on_mouse_enter", x, y, current_section=section
                    )
            if prevent_dispatch_origin is EVENT_UNHANDLED:
                prevent_dispatch_origin = self.dispatch_mouse_event(
                    event_origin, x, y, *args, **kwargs
                )

        # at the end catch the sections the mouse is moving over
        self.mouse_over_sections = current_sections

        # NOTE: the result from mouse enter/leave events is ignored here
        return prevent_dispatch_origin

    def on_mouse_motion(self, x: int, y: int, *args, **kwargs) -> bool | None:
        """
        This method dispatches the on_mouse_motion and also calculates if
        on_mouse_enter/leave should be fired.

        Args:
            x: the x axis coordinate
            y: the y axis coordinate
            args: any other position arguments that should be delivered to the dispatched event
            kwargs: any other keyword arguments that should be delivered to the dispatched event
        Returns:
            ``EVENT_HANDLED`` or ``EVENT_UNHANDLED``, or whatever the dispatched method returns
        """
        return self.dispatch_mouse_enter_leave_events("on_mouse_motion", x, y, *args, **kwargs)

    def on_mouse_drag(self, x: int, y: int, *args, **kwargs) -> bool | None:
        """
        This method dispatches the on_mouse_drag and also calculates if
        on_mouse_enter/leave should be fired.

        Args:
            x: the x axis coordinate
            y: the y axis coordinate
            args: any other position arguments that should be delivered to the dispatched event
            kwargs: any other keyword arguments that should be delivered to the dispatched event
        Returns:
            ``EVENT_HANDLED`` or ``EVENT_UNHANDLED``, or whatever the dispatched method returns
        """
        return self.dispatch_mouse_enter_leave_events("on_mouse_drag", x, y, *args, **kwargs)

    def on_mouse_scroll(self, x: int, y: int, *args, **kwargs) -> bool | None:
        """
        Triggers the on_mouse_scroll event on the appropriate sections or view

        Args:
            x: the x axis coordinate
            y: the y axis coordinate
            args: any other position arguments that should be delivered to the dispatched event
            kwargs: any other keyword arguments that should be delivered to the dispatched event
        Returns:
            ``EVENT_HANDLED`` or ``EVENT_UNHANDLED``, or whatever the dispatched method returns
        """
        return self.dispatch_mouse_event("on_mouse_scroll", x, y, *args, **kwargs)

    def on_mouse_enter(self, x: int, y: int, *args, **kwargs) -> bool | None:
        """
        Triggered when the mouse enters the window space
        Will trigger on_mouse_enter on the appropriate sections or view

        Args:
            x: the x axis coordinate
            y: the y axis coordinate
            args: any other position arguments that should be delivered to the dispatched event
            kwargs: any other keyword arguments that should be delivered to the dispatched event
        Returns:
            ``EVENT_HANDLED`` or ``EVENT_UNHANDLED``, or whatever the dispatched method returns
        """
        current_sections = list(self.get_sections(x, y))  # consume the generator

        # set the sections the mouse is over
        self.mouse_over_sections = current_sections

        prevent_dispatch = EVENT_UNHANDLED
        for section in current_sections:
            if prevent_dispatch is EVENT_HANDLED:
                break
            prevent_dispatch = self.dispatch_mouse_event(
                "on_mouse_enter", x, y, *args, **kwargs, current_section=section
            )
        return prevent_dispatch

    def on_mouse_leave(self, x: int, y: int, *args, **kwargs) -> bool | None:
        """
        Triggered when the mouse leaves the window space
        Will trigger on_mouse_leave on the appropriate sections or view

        Args:
            x: the x axis coordinate
            y: the y axis coordinate
            args: any other position arguments that should be delivered to the dispatched event
            kwargs: any other keyword arguments that should be delivered to the dispatched event
        Returns:
            ``EVENT_HANDLED`` or ``EVENT_UNHANDLED``, or whatever the dispatched method returns
        """
        prevent_dispatch = EVENT_UNHANDLED
        for section in self.mouse_over_sections:
            if prevent_dispatch is EVENT_HANDLED:
                break
            prevent_dispatch = self.dispatch_mouse_event(
                "on_mouse_leave", x, y, *args, **kwargs, current_section=section
            )
        # clear the sections the mouse is over as it's out of the screen
        self.mouse_over_sections = []
        return prevent_dispatch

    def on_key_press(self, *args, **kwargs) -> bool | None:
        """
        Triggers the on_key_press event on the appropriate sections or view

        Args:
            args: any other position arguments that should be delivered to the dispatched event
            kwargs: any other keyword arguments that should be delivered to the dispatched event
        Returns:
            ``EVENT_HANDLED`` or ``EVENT_UNHANDLED``, or whatever the dispatched method returns
        """
        return self.dispatch_keyboard_event("on_key_press", *args, **kwargs)

    def on_key_release(self, *args, **kwargs) -> bool | None:
        """
        Triggers the on_key_release event on the appropriate sections or view

        Args:
            args: any other position arguments that should be delivered to the dispatched event
            kwargs: any other keyword arguments that should be delivered to the dispatched event
        Returns:
            ``EVENT_HANDLED`` or ``EVENT_UNHANDLED``, or whatever the dispatched method returns
        """
        return self.dispatch_keyboard_event("on_key_release", *args, **kwargs)



"""
Sprite Facing Left or Right
Face left or right depending on our direction

Simple program to show basic sprite usage.

Artwork from https://kenney.nl

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.sprite_face_left_or_right
"""

import arcade

SPRITE_SCALING = 0.5

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Sprite Face Left or Right Example"

MOVEMENT_SPEED = 5

# Index of textures, first element faces left, second faces right
TEXTURE_LEFT = 0
TEXTURE_RIGHT = 1


class Player(arcade.Sprite):

    def __init__(self, left_texture, right_texture):
        super().__init__(left_texture, scale=SPRITE_SCALING)
        self.textures.append(right_texture)

    def update(self, delta_time: float = 1/60):
        self.center_x += self.change_x
        self.center_y += self.change_y

        # Figure out if we should face left or right
        if self.change_x < 0:
            self.texture = self.textures[TEXTURE_LEFT]
        elif self.change_x > 0:
            self.texture = self.textures[TEXTURE_RIGHT]


class GameView(arcade.View):
    """
    Main application class.
    """

    def __init__(self):
        """
        Initializer
        """

        # Call the parent class initializer
        super().__init__()

        # Variables that will hold sprite lists
        self.player_sprite_list = None

        # Set up the player info
        self.player_sprite = None

        # Set the background color
        self.background_color = arcade.color.AMAZON

        # Textures for left and right facing sprites
        self.left_texture = arcade.load_texture(":resources:images/enemies/bee.png")
        self.right_texture = self.left_texture.flip_left_right()

    def setup(self):
        """ Set up the game and initialize the variables. """

        # Sprite lists
        self.player_sprite_list = arcade.SpriteList()

        # Set up the player
        self.player_sprite = Player(self.left_texture, self.right_texture)
        self.player_sprite.center_x = WINDOW_WIDTH / 2
        self.player_sprite.center_y = WINDOW_HEIGHT / 2
        self.player_sprite_list.append(self.player_sprite)

    def on_draw(self):
        """
        Render the screen.
        """

        # Clear the screen with the configured background color
        self.clear()

        # Draw all the sprites.
        self.player_sprite_list.draw()

    def on_update(self, delta_time):
        """ Movement and game logic """

        # Call update on all sprites (The sprites don't do much in this
        # example though.)
        self.player_sprite_list.update(delta_time)

    def on_key_press(self, key, modifiers):
        """Called whenever a key is pressed. """

        if key in (arcade.key.UP, arcade.key.W):
            self.player_sprite.change_y = MOVEMENT_SPEED
        elif key in (arcade.key.DOWN, arcade.key.S):
            self.player_sprite.change_y = -MOVEMENT_SPEED
        elif key in (arcade.key.LEFT, arcade.key.A):
            self.player_sprite.change_x = -MOVEMENT_SPEED
        elif key in (arcade.key.RIGHT, arcade.key.D):
            self.player_sprite.change_x = MOVEMENT_SPEED
        elif key == arcade.key.ESCAPE:
            arcade.close_window()

    def on_key_release(self, key, modifiers):
        """Called when the user releases a key. """

        if key in (arcade.key.UP, arcade.key.DOWN, arcade.key.W, arcade.key.S):
            self.player_sprite.change_y = 0
        elif key in (arcade.key.LEFT, arcade.key.RIGHT, arcade.key.A, arcade.key.D):
            self.player_sprite.change_x = 0


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()
    game.setup()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



from pyglet.math import Vec3

from arcade.camera.data_types import CameraData


def strafe(data: CameraData, direction: tuple[float, float]) -> tuple[float, float, float]:
    """
    Move the CameraData in a 2D direction aligned to the up-right plane of the view.
    A value of [1, 0] will move the camera sideways while a value of [0, 1]
    will move the camera upwards. Works irrespective of which direction the camera is facing.
    """
    _forward = Vec3(*data.forward).normalize()
    _up = Vec3(*data.up).normalize()
    _right = _forward.cross(_up)

    _pos = data.position

    offset = _right * direction[0] + _up * direction[1]
    # fmt: off
    return (
            _pos[0] + offset[0],
            _pos[1] + offset[1],
            _pos[2] + offset[2]
        )
    # fmt: on



"""
Sound Panning Demo

If Python and Arcade are installed, this example can be run from the
command line with:
python -m arcade.examples.sound_demo

Each button plays a sound when clicked.

The top left button plays a streaming music track when pressed. If you
click it while it's already playing, it will intentionally crash the
demo to demonstrate how you shouldn't try to play a streaming sound
that's already playing.

The lower 3 rows of buttons play a non-streaming (static) sound with
different panning and volume. Going from left to right changes the
panning, which is how much the sound plays in the left speaker vs the
right speaker. Lower rows play the sound louder than the higher ones.
"""

import typing

import arcade

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Sound Panning Demo"
BUTTON_SIZE = 30


SOUND_PANNING = [-1.0, -0.5, 0.0, 0.5, 1.0]
MARGIN = WINDOW_WIDTH / 4
BUTTON_X_POSITIONS = [
    MARGIN,
    MARGIN + (WINDOW_WIDTH - MARGIN * 2) / 3 * 1,
    MARGIN + (WINDOW_WIDTH - MARGIN * 2) / 3 * 2,
    MARGIN + (WINDOW_WIDTH - MARGIN * 2) / 3 * 3,
    WINDOW_WIDTH - MARGIN,
]


VOLUME_VARIATION = [0.1, 0.5, 1]
Y_OFFSETS = [50, 0, -50]


class SoundButton(arcade.SpriteSolidColor):
    """
    A sprite that stores settings about how to play a sound.

    This class can load a sound as either a static sound or a streaming
    sound. Streaming should be used for long files that will only have
    one instance playing, such as music or ambiance tracks.

    If you try to play a sound created with streaming=True while it is
    already playing, it will raise an exception! Non-streaming (static)
    sounds are fine with it, and can have play() called on them as many
    times as you want.
    """

    def __init__(
        self,
        sound_file,
        pan=0.5,
        volume=0.5,
        center_x=0,
        center_y=0,
        streaming=False
    ):
        super().__init__(BUTTON_SIZE, BUTTON_SIZE, color=arcade.color.WHITE)
        self.sound = arcade.Sound(sound_file, streaming=streaming)
        self.pan = pan
        self.volume = volume
        self.center_x = center_x
        self.center_y = center_y

    def play(self):
        self.sound.play(pan=self.pan, volume=self.volume)


class GameView(arcade.View):
    def __init__(self):
        super().__init__()
        self.background_color = arcade.color.AMAZON
        self.button_sprites = None

    def setup(self):
        self.button_sprites = arcade.SpriteList()

        # create the streaming button at the top left
        self.button_sprites.append(
            SoundButton(
                ":resources:music/funkyrobot.mp3",
                pan=-1.0,
                volume=0.1,
                center_x=BUTTON_X_POSITIONS[0],
                center_y=WINDOW_HEIGHT / 2 + 150,
                streaming=True
            )
        )

        # Position the grid of buttons
        # The zip function takes pieces from iterables and returns them
        # as tuples. For more information, see the python doc:
        # https://docs.python.org/3/library/functions.html#zip
        for vol, y_offset in zip(VOLUME_VARIATION, Y_OFFSETS):
            for pan_setting, x_pos in zip(SOUND_PANNING, BUTTON_X_POSITIONS):
                self.button_sprites.append(
                    SoundButton(
                        ":resources:sounds/upgrade4.wav",
                        pan_setting,
                        vol,
                        x_pos,
                        WINDOW_HEIGHT / 2 + y_offset,
                    )
                )

    def on_draw(self):
        self.clear()
        self.button_sprites.draw()

    def on_update(self, delta_time):
        self.button_sprites.update()

    def on_mouse_press(self, x, y, button, key_modifiers):
        hit_sprites = arcade.get_sprites_at_point((x, y), self.button_sprites)
        for sprite in hit_sprites:
            button_sprite = typing.cast(SoundButton, sprite)
            if button == arcade.MOUSE_BUTTON_LEFT:
                button_sprite.play()


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()
    game.setup()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



from __future__ import annotations

import weakref
from ctypes import byref, string_at
from typing import TYPE_CHECKING

from pyglet import gl

from arcade.types import BufferProtocol

from .utils import data_to_ctypes

if TYPE_CHECKING:
    from arcade.gl import Context


class Buffer:
    """OpenGL buffer object. Buffers store byte data and upload it
    to graphics memory so shader programs can process the data.
    They are used for storage of vertex data,
    element data (vertex indexing), uniform block data etc.

    The ``data`` parameter can be anything that implements the
    `Buffer Protocol <https://docs.python.org/3/c-api/buffer.html>`_.

    This includes ``bytes``, ``bytearray``, ``array.array``, and
    more. You may need to use typing workarounds for non-builtin
    types. See :ref:`prog-guide-gl-buffer-protocol-typing` for more
    information.

    .. warning:: Buffer objects should be created using :py:meth:`arcade.gl.Context.buffer`

    Args:
        ctx:
            The context this buffer belongs to
        data:
            The data this buffer should contain. It can be a ``bytes`` instance or any
            object supporting the buffer protocol.
        reserve:
            Create a buffer of a specific byte size
        usage:
            A hit of this buffer is ``static`` or ``dynamic`` (can mostly be ignored)
    """

    __slots__ = "_ctx", "_glo", "_size", "_usage", "__weakref__"
    _usages = {
        "static": gl.GL_STATIC_DRAW,
        "dynamic": gl.GL_DYNAMIC_DRAW,
        "stream": gl.GL_STREAM_DRAW,
    }

    def __init__(
        self,
        ctx: Context,
        data: BufferProtocol | None = None,
        reserve: int = 0,
        usage: str = "static",
    ):
        self._ctx = ctx
        self._glo = glo = gl.GLuint()
        self._size = -1
        self._usage = Buffer._usages[usage]

        gl.glGenBuffers(1, byref(self._glo))
        # print(f"glGenBuffers() -> {self._glo.value}")
        if self._glo.value == 0:
            raise RuntimeError("Cannot create Buffer object.")

        # print(f"glBindBuffer({self._glo.value})")
        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, self._glo)
        # print(f"glBufferData(gl.GL_ARRAY_BUFFER, {self._size}, data, {self._usage})")

        if data is not None and len(data) > 0:  # type: ignore
            self._size, data = data_to_ctypes(data)
            gl.glBufferData(gl.GL_ARRAY_BUFFER, self._size, data, self._usage)
        elif reserve > 0:
            self._size = reserve
            # populate the buffer with zero byte values
            data = (gl.GLubyte * self._size)()
            gl.glBufferData(gl.GL_ARRAY_BUFFER, self._size, data, self._usage)
        else:
            raise ValueError("Buffer takes byte data or number of reserved bytes")

        if self._ctx.gc_mode == "auto":
            weakref.finalize(self, Buffer.delete_glo, self.ctx, glo)

        self._ctx.stats.incr("buffer")

    def __repr__(self):
        return f"<Buffer {self._glo.value}>"

    def __del__(self):
        # Intercept garbage collection if we are using Context.gc()
        if self._ctx.gc_mode == "context_gc" and self._glo.value > 0:
            self._ctx.objects.append(self)

    @property
    def size(self) -> int:
        """The byte size of the buffer."""
        return self._size

    @property
    def ctx(self) -> "Context":
        """The context this resource belongs to."""
        return self._ctx

    @property
    def glo(self) -> gl.GLuint:
        """The OpenGL resource id."""
        return self._glo

    def delete(self) -> None:
        """
        Destroy the underlying OpenGL resource.

        .. warning:: Don't use this unless you know exactly what you are doing.
        """
        Buffer.delete_glo(self._ctx, self._glo)
        self._glo.value = 0

    @staticmethod
    def delete_glo(ctx: Context, glo: gl.GLuint):
        """
        Release/delete open gl buffer.

        This is automatically called when the object is garbage collected.

        Args:
            ctx:
                The context the buffer belongs to
            glo:
                The OpenGL buffer id
        """
        # If we have no context, then we are shutting down, so skip this
        if gl.current_context is None:
            return

        if glo.value != 0:
            gl.glDeleteBuffers(1, byref(glo))
            glo.value = 0

        ctx.stats.decr("buffer")

    def read(self, size: int = -1, offset: int = 0) -> bytes:
        """Read data from the buffer.

        Args:
            size:
                The bytes to read. -1 means the entire buffer (default)
            offset:
                Byte read offset
        """
        if size == -1:
            size = self._size - offset

        # Catch this before confusing INVALID_OPERATION is raised
        if size < 1:
            raise ValueError(
                "Attempting to read 0 or less bytes from buffer: "
                f"buffer size={self._size} | params: size={size}, offset={offset}"
            )

        # Manually detect this so it doesn't raise a confusing INVALID_VALUE error
        if size + offset > self._size:
            raise ValueError(
                (
                    "Attempting to read outside the buffer. "
                    f"Buffer size: {self._size} "
                    f"Reading from {offset} to {size + offset}"
                )
            )

        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, self._glo)
        ptr = gl.glMapBufferRange(gl.GL_ARRAY_BUFFER, offset, size, gl.GL_MAP_READ_BIT)
        data = string_at(ptr, size=size)
        gl.glUnmapBuffer(gl.GL_ARRAY_BUFFER)
        return data

    def write(self, data: BufferProtocol, offset: int = 0):
        """Write byte data to the buffer from a buffer protocol object.

        The ``data`` value can be anything that implements the
        `Buffer Protocol <https://docs.python.org/3/c-api/buffer.html>`_.

        This includes ``bytes``, ``bytearray``, ``array.array``, and
        more. You may need to use typing workarounds for non-builtin
        types. See :ref:`prog-guide-gl-buffer-protocol-typing` for more
        information.

        If the supplied data is larger than the buffer, it will be
        truncated to fit. If the supplied data is smaller than the
        buffer, the remaining bytes will be left unchanged.

        Args:
            data:
                The byte data to write. This can be bytes or any object
                supporting the buffer protocol.
            offset:
                The byte offset
        """
        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, self._glo)
        size, data = data_to_ctypes(data)
        # Ensure we don't write outside the buffer
        size = min(size, self._size - offset)
        if size < 0:
            raise ValueError("Attempting to write negative number bytes to buffer")
        gl.glBufferSubData(gl.GL_ARRAY_BUFFER, gl.GLintptr(offset), size, data)

    def copy_from_buffer(self, source: Buffer, size=-1, offset=0, source_offset=0):
        """Copy data into this buffer from another buffer.

        Args:
            source:
                The buffer to copy from
            size:
                The amount of bytes to copy
            offset:
                The byte offset to write the data in this buffer
            source_offset:
                The byte offset to read from the source buffer
        """
        # Read the entire source buffer into this buffer
        if size == -1:
            size = source.size

        # TODO: Check buffer bounds
        if size + source_offset > source.size:
            raise ValueError("Attempting to read outside the source buffer")

        if size + offset > self._size:
            raise ValueError("Attempting to write outside the buffer")

        gl.glBindBuffer(gl.GL_COPY_READ_BUFFER, source.glo)
        gl.glBindBuffer(gl.GL_COPY_WRITE_BUFFER, self._glo)
        gl.glCopyBufferSubData(
            gl.GL_COPY_READ_BUFFER,
            gl.GL_COPY_WRITE_BUFFER,
            gl.GLintptr(source_offset),  # readOffset
            gl.GLintptr(offset),  # writeOffset
            size,  # size (number of bytes to copy)
        )

    def orphan(self, size: int = -1, double: bool = False):
        """
        Re-allocate the entire buffer memory. This can be used to resize
        a buffer or for re-specification (orphan the buffer to avoid blocking).

        If the current buffer is busy in rendering operations
        it will be deallocated by OpenGL when completed.

        Args:
            size:
                New size of buffer. -1 will retain the current size.
                Takes precedence over ``double`` parameter if specified.
            double:
                Is passed in with `True` the buffer size will be doubled
                from its current size.
        """
        if size > 0:
            self._size = size
        elif double is True:
            self._size *= 2

        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, self._glo)
        gl.glBufferData(gl.GL_ARRAY_BUFFER, self._size, None, self._usage)

    def bind_to_uniform_block(self, binding: int = 0, offset: int = 0, size: int = -1):
        """Bind this buffer to a uniform block location.
        In most cases it will be sufficient to only provide a binding location.

        Args:
            binding:
                The binding location
            offset:
                Byte offset
            size:
                Size of the buffer to bind.
        """
        if size < 0:
            size = self.size

        gl.glBindBufferRange(gl.GL_UNIFORM_BUFFER, binding, self._glo, offset, size)

    def bind_to_storage_buffer(self, *, binding=0, offset=0, size=-1):
        """
        Bind this buffer as a shader storage buffer.

        Args:
            binding:
                The binding location
            offset:
                Byte offset in the buffer
            size:
                The size in bytes. The entire buffer will be mapped by default.
        """
        if size < 0:
            size = self.size

        gl.glBindBufferRange(gl.GL_SHADER_STORAGE_BUFFER, binding, self._glo, offset, size)



"""
Particle Fireworks

Use a fireworks display to demonstrate "real-world" uses of Emitters and Particles

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.particle_fireworks
"""
import random

import pyglet
from pyglet.math import Vec2

import arcade
from arcade.types import Point, PathOrTexture, LBWH
from arcade.math import rand_in_rect, clamp, lerp, rand_in_circle, rand_on_circle
from arcade.particles import (
    Emitter,
    LifetimeParticle,
    FadeParticle,
    EmitterIntervalWithTime,
    EmitMaintainCount,
    EmitBurst,
)

WINDOW_WIDTH = 800
WINDOW_HEIGHT = 600
WINDOW_TITLE = "Particle based fireworks"
LAUNCH_INTERVAL_MIN = 1.5
LAUNCH_INTERVAL_MAX = 2.5
TEXTURE = "images/pool_cue_ball.png"
RAINBOW_COLORS = (
    arcade.color.ELECTRIC_CRIMSON,
    arcade.color.FLUORESCENT_ORANGE,
    arcade.color.ELECTRIC_YELLOW,
    arcade.color.ELECTRIC_GREEN,
    arcade.color.ELECTRIC_CYAN,
    arcade.color.MEDIUM_ELECTRIC_BLUE,
    arcade.color.ELECTRIC_INDIGO,
    arcade.color.ELECTRIC_PURPLE,
)
SPARK_TEXTURES = [arcade.make_circle_texture(8, clr) for clr in RAINBOW_COLORS]
SPARK_PAIRS = [
    [SPARK_TEXTURES[0], SPARK_TEXTURES[3]],
    [SPARK_TEXTURES[1], SPARK_TEXTURES[5]],
    [SPARK_TEXTURES[7], SPARK_TEXTURES[2]],
]
ROCKET_SMOKE_TEXTURE = arcade.make_soft_circle_texture(15, arcade.color.GRAY)
PUFF_TEXTURE = arcade.make_soft_circle_texture(80, (40, 40, 40, 255))
FLASH_TEXTURE = arcade.make_soft_circle_texture(70, (128, 128, 90, 255))
CLOUD_TEXTURES = [
    arcade.make_soft_circle_texture(50, arcade.color.WHITE),
    arcade.make_soft_circle_texture(50, arcade.color.LIGHT_GRAY),
    arcade.make_soft_circle_texture(50, arcade.color.LIGHT_BLUE),
]
STAR_TEXTURES = [
    arcade.make_soft_circle_texture(8, arcade.color.WHITE),
    arcade.make_soft_circle_texture(8, arcade.color.PASTEL_YELLOW),
]
SPINNER_HEIGHT = 75


def make_spinner():
    spinner = Emitter(
        center_xy=(WINDOW_WIDTH / 2, SPINNER_HEIGHT - 5),
        emit_controller=EmitterIntervalWithTime(0.025, 2.0),
        particle_factory=lambda emitter: FadeParticle(
            filename_or_texture=random.choice(STAR_TEXTURES), change_xy=(0, 6.0), lifetime=0.2
        ),
    )
    spinner.change_angle = 16.28
    return spinner


def make_rocket(emit_done_cb):
    """Emitter that displays the smoke trail as the firework shell climbs into the sky"""
    rocket = RocketEmitter(
        center_xy=(random.uniform(100, WINDOW_WIDTH - 100), 25),
        emit_controller=EmitterIntervalWithTime(0.04, 2.0),
        particle_factory=lambda emitter: FadeParticle(
            filename_or_texture=ROCKET_SMOKE_TEXTURE,
            change_xy=rand_in_circle((0.0, 0.0), 0.08),
            scale=0.5,
            lifetime=random.uniform(1.0, 1.5),
            start_alpha=100,
            end_alpha=0,
            mutation_callback=rocket_smoke_mutator,
        ),
        emit_done_cb=emit_done_cb,
    )
    rocket.change_x = random.uniform(-1.0, 1.0)
    rocket.change_y = random.uniform(5.0, 7.25)
    return rocket


def make_flash(prev_emitter):
    """Return emitter that displays the brief flash when a firework shell explodes"""
    return Emitter(
        center_xy=prev_emitter.get_pos(),
        emit_controller=EmitBurst(3),
        particle_factory=lambda emitter: FadeParticle(
            filename_or_texture=FLASH_TEXTURE,
            change_xy=rand_in_circle((0.0, 0.0), 3.5),
            lifetime=0.15,
        ),
    )


def make_puff(prev_emitter):
    """
    Return emitter that generates the subtle smoke cloud
    left after a firework shell explodes.
    """
    return Emitter(
        center_xy=prev_emitter.get_pos(),
        emit_controller=EmitBurst(4),
        particle_factory=lambda emitter: FadeParticle(
            filename_or_texture=PUFF_TEXTURE,
            change_xy=Vec2(*rand_in_circle((0.0, 0.0), 0.4)) + Vec2(0.3, 0.0),
            lifetime=4.0,
        ),
    )


class AnimatedAlphaParticle(LifetimeParticle):
    """A custom particle that animates between three different alpha levels"""

    def __init__(
        self,
        filename_or_texture: PathOrTexture | None,
        change_xy: Vec2,
        start_alpha: int = 0,
        duration1: float = 1.0,
        mid_alpha: int = 255,
        duration2: float = 1.0,
        end_alpha: int = 0,
        center_xy: Point = (0.0, 0.0),
        angle: float = 0,
        change_angle: float = 0,
        scale: float = 1.0,
        mutation_callback=None,
    ):
        super().__init__(
            filename_or_texture,
            change_xy,
            duration1 + duration2,
            center_xy,
            angle,
            change_angle,
            scale,
            start_alpha,
            mutation_callback,
        )
        self.start_alpha = start_alpha
        self.in_duration = duration1
        self.mid_alpha = mid_alpha
        self.out_duration = duration2
        self.end_alpha = end_alpha

    def update(self, delta_time: float = 1 / 60):
        super().update(delta_time)
        if self.lifetime_elapsed <= self.in_duration:
            u = self.lifetime_elapsed / self.in_duration
            self.alpha = int(clamp(lerp(self.start_alpha, self.mid_alpha, u), 0, 255))
        else:
            u = (self.lifetime_elapsed - self.in_duration) / self.out_duration
            self.alpha = int(clamp(lerp(self.mid_alpha, self.end_alpha, u), 0, 255))


class RocketEmitter(Emitter):
    """
    Custom emitter class to add gravity to the emitter to
    represent gravity on the firework shell.
    """
    def update(self, delta_time: float = 1 / 60):
        super().update(delta_time)
        # gravity
        self.change_y += -0.05 * (60 * delta_time)


class GameView(arcade.View):
    def __init__(self):
        super().__init__()

        self.background_color = arcade.color.BLACK
        self.emitters: list[Emitter] = []

        self.launch_firework(0)
        arcade.schedule(self.launch_spinner, 4.0)

        stars = Emitter(
            center_xy=(0.0, 0.0),
            emit_controller=EmitMaintainCount(20),
            particle_factory=lambda emitter: AnimatedAlphaParticle(
                filename_or_texture=random.choice(STAR_TEXTURES),
                change_xy=(0.0, 0.0),
                start_alpha=0,
                duration1=random.uniform(2.0, 6.0),
                mid_alpha=128,
                duration2=random.uniform(2.0, 6.0),
                end_alpha=0,
                center_xy=rand_in_rect(LBWH(0.0, 0.0, WINDOW_WIDTH, WINDOW_HEIGHT)),
            ),
        )
        self.emitters.append(stars)

        x, y = rand_in_circle(center=(0.0, 0.0), radius=0.04)
        change_vec2 = Vec2(x, y) + Vec2(0.1, 0)
        change_tuple = change_vec2.x, change_vec2.y
        self.cloud = Emitter(
            center_xy=(50, 500),
            change_xy=(0.15, 0),
            emit_controller=EmitMaintainCount(60),
            particle_factory=lambda emitter: AnimatedAlphaParticle(
                filename_or_texture=random.choice(CLOUD_TEXTURES),
                change_xy=change_tuple,
                start_alpha=0,
                duration1=random.uniform(5.0, 10.0),
                mid_alpha=255,
                duration2=random.uniform(5.0, 10.0),
                end_alpha=0,
                center_xy=rand_in_circle((0.0, 0.0), 50),
            ),
        )
        self.emitters.append(self.cloud)

    def launch_firework(self, delta_time):
        launchers = (
            self.launch_random_firework,
            self.launch_ringed_firework,
            self.launch_sparkle_firework,
        )
        random.choice(launchers)(delta_time)
        pyglet.clock.schedule_once(
            self.launch_firework,
            random.uniform(LAUNCH_INTERVAL_MIN, LAUNCH_INTERVAL_MAX),
        )

    def launch_random_firework(self, _delta_time):
        """Simple firework that explodes in a random color"""
        rocket = make_rocket(self.explode_firework)
        self.emitters.append(rocket)

    def launch_ringed_firework(self, _delta_time):
        """ "Firework that has a basic explosion and a ring of sparks of a different color"""
        rocket = make_rocket(self.explode_ringed_firework)
        self.emitters.append(rocket)

    def launch_sparkle_firework(self, _delta_time):
        """Firework which has sparks that sparkle"""
        rocket = make_rocket(self.explode_sparkle_firework)
        self.emitters.append(rocket)

    def launch_spinner(self, _delta_time):
        """Start the spinner that throws sparks"""
        spinner1 = make_spinner()
        spinner2 = make_spinner()
        spinner2.angle = 180
        self.emitters.append(spinner1)
        self.emitters.append(spinner2)

    def explode_firework(self, prev_emitter):
        """
        Actions that happen when a firework shell explodes,
        resulting in a typical firework
        """
        self.emitters.append(make_puff(prev_emitter))
        self.emitters.append(make_flash(prev_emitter))

        spark_texture = random.choice(SPARK_TEXTURES)
        sparks = Emitter(
            center_xy=prev_emitter.get_pos(),
            emit_controller=EmitBurst(random.randint(30, 40)),
            particle_factory=lambda emitter: FadeParticle(
                filename_or_texture=spark_texture,
                change_xy=rand_in_circle(center=(0.0, 0.0), radius=9.0),
                lifetime=random.uniform(0.5, 1.2),
                mutation_callback=firework_spark_mutator,
            ),
        )
        self.emitters.append(sparks)

    def explode_ringed_firework(self, prev_emitter):
        """
        Actions that happen when a firework shell explodes,
        resulting in a ringed firework.
        """
        self.emitters.append(make_puff(prev_emitter))
        self.emitters.append(make_flash(prev_emitter))

        spark_texture, ring_texture = random.choice(SPARK_PAIRS)
        sparks = Emitter(
            center_xy=prev_emitter.get_pos(),
            emit_controller=EmitBurst(25),
            particle_factory=lambda emitter: FadeParticle(
                filename_or_texture=spark_texture,
                change_xy=rand_in_circle((0.0, 0.0), 8.0),
                lifetime=random.uniform(0.55, 0.8),
                mutation_callback=firework_spark_mutator,
            ),
        )
        self.emitters.append(sparks)

        ring = Emitter(
            center_xy=prev_emitter.get_pos(),
            emit_controller=EmitBurst(20),
            particle_factory=lambda emitter: FadeParticle(
                filename_or_texture=ring_texture,
                change_xy=rand_on_circle(center=(0.0, 0.0), radius=5.0),
                lifetime=random.uniform(1.0, 1.6),
                mutation_callback=firework_spark_mutator,
            ),
        )
        self.emitters.append(ring)

    def explode_sparkle_firework(self, prev_emitter):
        """
        Actions that happen when a firework shell explodes,
        resulting in a sparkling firework.
        """
        self.emitters.append(make_puff(prev_emitter))
        self.emitters.append(make_flash(prev_emitter))

        spark_texture = random.choice(SPARK_TEXTURES)
        sparks = Emitter(
            center_xy=prev_emitter.get_pos(),
            emit_controller=EmitBurst(random.randint(30, 40)),
            particle_factory=lambda emitter: AnimatedAlphaParticle(
                filename_or_texture=spark_texture,
                change_xy=rand_in_circle(center=(0.0, 0.0), radius=9.0),
                start_alpha=255,
                duration1=random.uniform(0.6, 1.0),
                mid_alpha=0,
                duration2=random.uniform(0.1, 0.2),
                end_alpha=255,
                mutation_callback=firework_spark_mutator,
            ),
        )
        self.emitters.append(sparks)

    def on_update(self, delta_time):
        # prevent list from being mutated (often by callbacks) while iterating over it
        emitters_to_update = self.emitters.copy()
        # update cloud
        if self.cloud.center_x > WINDOW_WIDTH:
            self.cloud.center_x = 0
        # update
        for e in emitters_to_update:
            e.update(delta_time)
        # remove emitters that can be reaped
        to_del = [e for e in emitters_to_update if e.can_reap()]
        for e in to_del:
            self.emitters.remove(e)

    def on_draw(self):
        self.clear()
        for e in self.emitters:
            e.draw()
        arcade.draw_lrbt_rectangle_filled(0, WINDOW_WIDTH, 0, 25, arcade.color.DARK_GREEN)
        mid = WINDOW_WIDTH / 2
        arcade.draw_lrbt_rectangle_filled(
            mid - 2,
            mid + 2,
            10,
            SPINNER_HEIGHT,
            arcade.color.DARK_BROWN,
        )

    def on_key_press(self, key, modifiers):
        if key == arcade.key.ESCAPE:
            self.window.close()


def firework_spark_mutator(particle: FadeParticle):
    """mutation_callback shared by all fireworks sparks"""
    # gravity
    particle.change_y += -0.03
    # drag
    particle.change_x *= 0.92
    particle.change_y *= 0.92


def rocket_smoke_mutator(particle: LifetimeParticle):
    particle.scale = lerp(
        0.5,
        3.0,
        particle.lifetime_elapsed / particle.lifetime_original,  # type: ignore
    )



def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""
Show how to do enemies in a platformer

Artwork from: https://kenney.nl
Tiled available from: https://www.mapeditor.org/

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.sprite_enemies_in_platformer
"""

import arcade

SPRITE_SCALING = 0.5
SPRITE_NATIVE_SIZE = 128
SPRITE_SIZE = int(SPRITE_NATIVE_SIZE * SPRITE_SCALING)

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Sprite Enemies in a Platformer Example"

# How many pixels to keep as a minimum margin between the character
# and the edge of the screen.
VIEWPORT_MARGIN = 40
RIGHT_MARGIN = 150

# Physics
MOVEMENT_SPEED = 5
JUMP_SPEED = 14
GRAVITY = 0.5


class GameView(arcade.View):
    """Main application class."""

    def __init__(self):
        """Initializer"""
        super().__init__()

        # Sprite lists
        self.wall_list = None
        self.enemy_list = None
        self.player_list = None

        # Set up the player
        self.player_sprite = None
        self.physics_engine = None
        self.view_left = 0
        self.view_bottom = 0
        self.game_over = False

    def setup(self):
        """Set up the game and initialize the variables."""
        # Sprite lists
        self.wall_list = arcade.SpriteList()
        self.enemy_list = arcade.SpriteList()
        self.player_list = arcade.SpriteList()

        # Draw the walls on the bottom
        for x in range(0, WINDOW_WIDTH, SPRITE_SIZE):
            wall = arcade.Sprite(
                ":resources:images/tiles/grassMid.png",
                scale=SPRITE_SCALING,
            )

            wall.bottom = 0
            wall.left = x
            self.wall_list.append(wall)

        # Draw the platform
        for x in range(SPRITE_SIZE * 3, SPRITE_SIZE * 8, SPRITE_SIZE):
            wall = arcade.Sprite(
                ":resources:images/tiles/grassMid.png",
                scale=SPRITE_SCALING,
            )

            wall.bottom = SPRITE_SIZE * 3
            wall.left = x
            self.wall_list.append(wall)

        # Draw the crates
        for x in range(0, WINDOW_WIDTH, SPRITE_SIZE * 5):
            wall = arcade.Sprite(
                ":resources:images/tiles/boxCrate_double.png",
                scale=SPRITE_SCALING,
            )

            wall.bottom = SPRITE_SIZE
            wall.left = x
            self.wall_list.append(wall)

        # -- Draw an enemy on the ground
        enemy = arcade.Sprite(
            ":resources:images/enemies/wormGreen.png",
            scale=SPRITE_SCALING,
        )

        enemy.bottom = SPRITE_SIZE
        enemy.left = SPRITE_SIZE * 2

        # Set enemy initial speed
        enemy.change_x = 2
        self.enemy_list.append(enemy)

        # -- Draw a enemy on the platform
        enemy = arcade.Sprite(
            ":resources:images/enemies/wormGreen.png",
            scale=SPRITE_SCALING,
        )

        enemy.bottom = SPRITE_SIZE * 4
        enemy.left = SPRITE_SIZE * 4

        # Set boundaries on the left/right the enemy can't cross
        enemy.boundary_right = SPRITE_SIZE * 8
        enemy.boundary_left = SPRITE_SIZE * 3
        enemy.change_x = 2
        self.enemy_list.append(enemy)

        # -- Set up the player
        self.player_sprite = arcade.Sprite(
            ":resources:images/animated_characters/female_person/femalePerson_idle.png",
            scale=SPRITE_SCALING,
        )
        self.player_list.append(self.player_sprite)

        # Starting position of the player
        self.player_sprite.center_x = 64
        self.player_sprite.center_y = 270

        self.physics_engine = arcade.PhysicsEnginePlatformer(
            self.player_sprite,
            self.wall_list,
            gravity_constant=GRAVITY,
        )

        # Set the background color
        self.background_color = arcade.color.AMAZON

    def on_draw(self):
        """
        Render the screen.
        """

        # This command has to happen before we start drawing
        self.clear()

        # Draw all the sprites.
        self.player_list.draw()
        self.wall_list.draw()
        self.enemy_list.draw()

        if self.game_over:
            arcade.draw_text(
                "GAME OVER, Press R to reset",
                self.center_x, self.center_y,
                font_size=30, anchor_x='center'
            )

    def on_key_press(self, key, modifiers):
        """
        Called whenever the mouse moves.
        """
        if key == arcade.key.UP:
            if self.physics_engine.can_jump():
                self.player_sprite.change_y = JUMP_SPEED
        elif key == arcade.key.LEFT:
            self.player_sprite.change_x = -MOVEMENT_SPEED
        elif key == arcade.key.RIGHT:
            self.player_sprite.change_x = MOVEMENT_SPEED
        elif key == arcade.key.R:
            self.setup()
            self.game_over = False

    def on_key_release(self, key, modifiers):
        """
        Called when the user presses a mouse button.
        """
        if key == arcade.key.LEFT or key == arcade.key.RIGHT:
            self.player_sprite.change_x = 0

    def on_update(self, delta_time):
        """ Movement and game logic """

        # Update the player based on the physics engine
        if not self.game_over:
            # Move the enemies
            self.enemy_list.update()

            # Check each enemy
            for enemy in self.enemy_list:
                # If the enemy hit a wall, reverse
                if len(arcade.check_for_collision_with_list(enemy, self.wall_list)) > 0:
                    enemy.change_x *= -1
                # If the enemy hit the left boundary, reverse
                elif enemy.boundary_left is not None and enemy.left < enemy.boundary_left:
                    enemy.change_x *= -1
                # If the enemy hit the right boundary, reverse
                elif enemy.boundary_right is not None and enemy.right > enemy.boundary_right:
                    enemy.change_x *= -1

            # Update the player using the physics engine
            self.physics_engine.update()

            # See if the player hit a worm. If so, game over.
            if len(arcade.check_for_collision_with_list(
                self.player_sprite,
                self.enemy_list,
            )) > 0:
                self.game_over = True


def main():
    """ Main function """
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)
    game = GameView()
    game.setup()

    window.show_view(game)
    arcade.run()


if __name__ == "__main__":
    main()



from __future__ import annotations

import logging
import weakref
from collections import deque
from contextlib import contextmanager
from ctypes import c_char_p, c_float, c_int, cast
from typing import (
    Any,
    Deque,
    Dict,
    Iterable,
    List,
    Literal,
    Sequence,
    Set,
    Tuple,
    overload,
)

import pyglet
import pyglet.gl.lib
from pyglet import gl
from pyglet.window import Window

from ..types import BufferProtocol
from .buffer import Buffer
from .compute_shader import ComputeShader
from .framebuffer import DefaultFrameBuffer, Framebuffer
from .glsl import ShaderSource
from .program import Program
from .query import Query
from .sampler import Sampler
from .texture import Texture2D
from .texture_array import TextureArray
from .types import BufferDescription, GLenumLike, PyGLenum
from .vertex_array import Geometry

LOG = logging.getLogger(__name__)


class Context:
    """
    Represents an OpenGL context. This context belongs to a pyglet window.
    normally accessed through ``window.ctx``.

    The Context class contains methods for creating resources,
    global states and commonly used enums. All enums also exist
    in the ``gl`` module. (``ctx.BLEND`` or ``arcade.gl.BLEND``).

    Args:
        window: The pyglet window this context belongs to
        gc_mode: The garbage collection mode. Default is "context_gc"
        gl_api: The OpenGL api. Default is "gl"
    """

    active: Context | None = None
    """The active context"""

    #: The OpenGL api. Usually "gl" or "gles".
    gl_api: str = "gl"

    # --- Store the most commonly used OpenGL constants
    # Texture

    NEAREST = 0x2600
    """Texture interpolation - Nearest pixel"""

    LINEAR = 0x2601
    """Texture interpolation - Linear interpolate"""

    NEAREST_MIPMAP_NEAREST = 0x2700
    """Texture interpolation - Minification filter for mipmaps"""

    LINEAR_MIPMAP_NEAREST = 0x2701
    """Texture interpolation - Minification filter for mipmaps"""

    NEAREST_MIPMAP_LINEAR = 0x2702
    """Texture interpolation - Minification filter for mipmaps"""

    LINEAR_MIPMAP_LINEAR = 0x2703
    """Texture interpolation - Minification filter for mipmaps"""

    REPEAT = gl.GL_REPEAT
    """Texture wrap mode - Repeat"""

    CLAMP_TO_EDGE = gl.GL_CLAMP_TO_EDGE
    """Texture wrap mode - Clamp to border pixel"""

    CLAMP_TO_BORDER = gl.GL_CLAMP_TO_BORDER
    """Texture wrap mode - Clamp to border color"""

    MIRRORED_REPEAT = gl.GL_MIRRORED_REPEAT
    """Texture wrap mode - Repeat mirrored"""

    # Flags

    BLEND = gl.GL_BLEND
    """Context flag - Blending"""

    DEPTH_TEST = gl.GL_DEPTH_TEST
    """Context flag - Depth testing"""

    CULL_FACE = gl.GL_CULL_FACE
    """Context flag - Face culling"""

    PROGRAM_POINT_SIZE = gl.GL_PROGRAM_POINT_SIZE
    """
    Context flag - Enables ``gl_PointSize`` in vertex or geometry shaders.

    When enabled we can write to ``gl_PointSize`` in the vertex shader to specify the point size
    for each individual point.

    If this value is not set in the shader the behavior is undefined. This means the points may
    or may not appear depending if the drivers enforce some default value for ``gl_PointSize``.

    When disabled :py:attr:`point_size` is used.
    """

    # Blend functions
    ZERO = 0x0000
    """Blend function"""

    ONE = 0x0001
    """Blend function"""

    SRC_COLOR = 0x0300
    """Blend function"""

    ONE_MINUS_SRC_COLOR = 0x0301
    """Blend function"""

    SRC_ALPHA = 0x0302
    """Blend function"""

    ONE_MINUS_SRC_ALPHA = 0x0303
    """Blend function"""

    DST_ALPHA = 0x0304
    """Blend function"""

    ONE_MINUS_DST_ALPHA = 0x0305
    """Blend function"""

    DST_COLOR = 0x0306
    """Blend function"""

    ONE_MINUS_DST_COLOR = 0x0307
    """Blend function"""

    # Blend equations
    FUNC_ADD = 0x8006
    """Blend equation - source + destination"""

    FUNC_SUBTRACT = 0x800A
    """Blend equation - source - destination"""

    FUNC_REVERSE_SUBTRACT = 0x800B
    """Blend equation - destination - source"""

    MIN = 0x8007
    """Blend equation - Minimum of source and destination"""

    MAX = 0x8008
    """Blend equation - Maximum of source and destination"""

    # Blend mode shortcuts
    BLEND_DEFAULT = 0x0302, 0x0303
    """Blend mode shortcut for default blend mode - ``SRC_ALPHA, ONE_MINUS_SRC_ALPHA``"""

    BLEND_ADDITIVE = 0x0001, 0x0001
    """Blend mode shortcut for additive blending - ``ONE, ONE``"""

    BLEND_PREMULTIPLIED_ALPHA = 0x0302, 0x0001
    """Blend mode shortcut for pre-multiplied alpha - ``SRC_ALPHA, ONE``"""

    # VertexArray: Primitives
    POINTS = gl.GL_POINTS  # 0
    """Primitive mode - points"""

    LINES = gl.GL_LINES  # 1
    """Primitive mode - lines"""

    LINE_LOOP = gl.GL_LINE_LOOP  # 2
    """Primitive mode - line loop"""

    LINE_STRIP = gl.GL_LINE_STRIP  # 3
    """Primitive mode - line strip"""

    TRIANGLES = gl.GL_TRIANGLES  # 4
    """Primitive mode - triangles"""

    TRIANGLE_STRIP = gl.GL_TRIANGLE_STRIP  # 5
    """Primitive mode - triangle strip"""

    TRIANGLE_FAN = gl.GL_TRIANGLE_FAN  # 6
    """Primitive mode - triangle fan"""

    LINES_ADJACENCY = gl.GL_LINES_ADJACENCY  # 10
    """Primitive mode - lines with adjacency"""

    LINE_STRIP_ADJACENCY = gl.GL_LINE_STRIP_ADJACENCY  # 11
    """Primitive mode - line strip with adjacency"""

    TRIANGLES_ADJACENCY = gl.GL_TRIANGLES_ADJACENCY  # 12
    """Primitive mode - triangles with adjacency"""

    TRIANGLE_STRIP_ADJACENCY = gl.GL_TRIANGLE_STRIP_ADJACENCY  # 13
    """Primitive mode - triangle strip with adjacency"""

    PATCHES = gl.GL_PATCHES
    """Primitive mode - Patch (tessellation)"""

    # The most common error enums
    _errors = {
        gl.GL_INVALID_ENUM: "GL_INVALID_ENUM",
        gl.GL_INVALID_VALUE: "GL_INVALID_VALUE",
        gl.GL_INVALID_OPERATION: "GL_INVALID_OPERATION",
        gl.GL_INVALID_FRAMEBUFFER_OPERATION: "GL_INVALID_FRAMEBUFFER_OPERATION",
        gl.GL_OUT_OF_MEMORY: "GL_OUT_OF_MEMORY",
        gl.GL_STACK_UNDERFLOW: "GL_STACK_UNDERFLOW",
        gl.GL_STACK_OVERFLOW: "GL_STACK_OVERFLOW",
    }
    _valid_apis = ("gl", "gles")

    def __init__(
        self,
        window: pyglet.window.Window,  # type: ignore
        gc_mode: str = "context_gc",
        gl_api: str = "gl",
    ):
        self._window_ref = weakref.ref(window)
        if gl_api not in self._valid_apis:
            raise ValueError(f"Invalid gl_api. Options are: {self._valid_apis}")
        self.gl_api = gl_api
        self._info = GLInfo(self)
        self._gl_version = (self._info.MAJOR_VERSION, self._info.MINOR_VERSION)
        Context.activate(self)
        # Texture unit we use when doing operations on textures to avoid
        # affecting currently bound textures in the first units
        self.default_texture_unit: int = self._info.MAX_TEXTURE_IMAGE_UNITS - 1

        # Detect the default framebuffer
        self._screen = DefaultFrameBuffer(self)
        # Tracking active program
        self.active_program: Program | ComputeShader | None = None
        # Tracking active framebuffer. On context creation the window is the default render target
        self.active_framebuffer: Framebuffer = self._screen
        self._stats: ContextStats = ContextStats(warn_threshold=1000)

        # Hardcoded states
        # This should always be enabled
        # gl.glEnable(gl.GL_TEXTURE_CUBE_MAP_SEAMLESS)
        # Set primitive restart index to -1 by default
        if self.gl_api == "gles":
            gl.glEnable(gl.GL_PRIMITIVE_RESTART_FIXED_INDEX)
        else:
            gl.glEnable(gl.GL_PRIMITIVE_RESTART)

        self._primitive_restart_index = -1
        self.primitive_restart_index = self._primitive_restart_index

        # Detect support for glProgramUniform.
        # Assumed to be supported in gles
        self._ext_separate_shader_objects_enabled = True
        if self.gl_api == "gl":
            have_ext = gl.gl_info.have_extension("GL_ARB_separate_shader_objects")
            self._ext_separate_shader_objects_enabled = self.gl_version >= (4, 1) or have_ext

        # We enable scissor testing by default.
        # This is always set to the same value as the viewport
        # to avoid background color affecting areas outside the viewport
        gl.glEnable(gl.GL_SCISSOR_TEST)

        # States
        self._blend_func: Tuple[int, int] | Tuple[int, int, int, int] = self.BLEND_DEFAULT
        self._point_size = 1.0
        self._flags: Set[int] = set()
        self._wireframe = False
        # Options for cull_face
        self._cull_face_options = {
            "front": gl.GL_FRONT,
            "back": gl.GL_BACK,
            "front_and_back": gl.GL_FRONT_AND_BACK,
        }
        self._cull_face_options_reverse = {
            gl.GL_FRONT: "front",
            gl.GL_BACK: "back",
            gl.GL_FRONT_AND_BACK: "front_and_back",
        }

        # Context GC as default. We need to call Context.gc() to free opengl resources
        self._gc_mode = "context_gc"
        self.gc_mode = gc_mode
        #: Collected objects to gc when gc_mode is "context_gc".
        #: This can be used during debugging.
        self.objects: Deque[Any] = deque()

    @property
    def info(self) -> GLInfo:
        """
        Get the info object for this context containing information
        about hardware/driver limits and other information.

        Example::

            >> ctx.info.MAX_TEXTURE_SIZE
            (16384, 16384)
            >> ctx.info.VENDOR
            NVIDIA Corporation
            >> ctx.info.RENDERER
            NVIDIA GeForce RTX 2080 SUPER/PCIe/SSE2
        """
        return self._info

    @property
    def extensions(self) -> set[str]:
        """
        Get a set of supported OpenGL extensions strings for this context.

        This can be used to check if a specific extension is supported::

            # Check if bindless textures are supported
            "GL_ARB_bindless_texture" in ctx.extensions
            # Check for multiple extensions
            expected_extensions = {"GL_ARB_bindless_texture", "GL_ARB_get_program_binary"}
            ctx.extensions & expected_extensions == expected_extensions
        """
        return gl.gl_info.get_extensions()

    @property
    def stats(self) -> ContextStats:
        """
        Get the stats instance containing runtime information
        about creation and destruction of OpenGL objects.

        This can be useful for debugging and profiling.
        Creating and throwing away OpenGL objects can be detrimental
        to performance.

        Example::

            # Show the created and freed resource count
            >> ctx.stats.texture
            (100, 10)
            >> ctx.framebuffer
            (1, 0)
            >> ctx.buffer
            (10, 0)
        """
        return self._stats

    @property
    def window(self) -> Window:
        """The window this context belongs to (read only)."""
        window_ref = self._window_ref()
        if window_ref is None:
            raise Exception("Window not available, lost reference.")
        return window_ref

    @property
    def screen(self) -> Framebuffer:
        """The framebuffer for the window (read only)"""
        return self._screen

    @property
    def fbo(self) -> Framebuffer:
        """
        Get the currently active framebuffer (read only).
        """
        return self.active_framebuffer

    @property
    def gl_version(self) -> Tuple[int, int]:
        """
        The OpenGL major and minor version as a tuple.

        This is the reported OpenGL version from
        drivers and might be a higher version than
        you requested.
        """
        return self._gl_version

    def gc(self) -> int:
        """
        Run garbage collection of OpenGL objects for this context.
        This is only needed when ``gc_mode`` is ``context_gc``.

        Returns:
            The number of resources destroyed
        """
        # Loop the array until all objects are gone.
        # Deleting one object might add new ones so we need
        # to loop until the deque is empty
        num_objects = 0

        while len(self.objects):
            obj = self.objects.popleft()
            obj.delete()
            num_objects += 1

        return num_objects

    @property
    def gc_mode(self) -> str:
        """
        Set the garbage collection mode for OpenGL resources.
        Supported modes are::

            # Default:
            # Defer garbage collection until ctx.gc() is called
            # This can be useful to enforce the main thread to
            # run garbage collection of opengl resources
            ctx.gc_mode = "context_gc"

            # Auto collect is similar to python garbage collection.
            # This is a risky mode. Know what you are doing before using this.
            ctx.gc_mode = "auto"
        """
        return self._gc_mode

    @gc_mode.setter
    def gc_mode(self, value: str):
        modes = ["auto", "context_gc"]
        if value not in modes:
            raise ValueError("Unsupported gc_mode. Supported modes are:", modes)
        self._gc_mode = value

    @property
    def error(self) -> str | None:
        """Check OpenGL error

        Returns a string representation of the occurring error
        or ``None`` of no errors has occurred.

        Example::

            err = ctx.error
            if err:
                raise RuntimeError("OpenGL error: {err}")
        """
        err = gl.glGetError()
        if err == gl.GL_NO_ERROR:
            return None

        return self._errors.get(err, "GL_UNKNOWN_ERROR")

    @classmethod
    def activate(cls, ctx: Context):
        """
        Mark a context as the currently active one.

        .. Warning:: Never call this unless you know exactly what you are doing.

        Args:
            ctx: The context to activate
        """
        cls.active = ctx

    def enable(self, *flags: int):
        """
        Enables one or more context flags::

            # Single flag
            ctx.enable(ctx.BLEND)
            # Multiple flags
            ctx.enable(ctx.DEPTH_TEST, ctx.CULL_FACE)

        Args:
            *flags: The flags to enable
        """
        self._flags.update(flags)

        for flag in flags:
            gl.glEnable(flag)

    def enable_only(self, *args: int):
        """
        Enable only some flags. This will disable all other flags.
        This is a simple way to ensure that context flag states
        are not lingering from other sections of your code base::

            # Ensure all flags are disabled (enable no flags)
            ctx.enable_only()
            # Make sure only blending is enabled
            ctx.enable_only(ctx.BLEND)
            # Make sure only depth test and culling is enabled
            ctx.enable_only(ctx.DEPTH_TEST, ctx.CULL_FACE)

        Args:
            *args: The flags to enable
        """
        self._flags = set(args)

        if self.BLEND in self._flags:
            gl.glEnable(self.BLEND)
        else:
            gl.glDisable(self.BLEND)

        if self.DEPTH_TEST in self._flags:
            gl.glEnable(self.DEPTH_TEST)
        else:
            gl.glDisable(self.DEPTH_TEST)

        if self.CULL_FACE in self._flags:
            gl.glEnable(self.CULL_FACE)
        else:
            gl.glDisable(self.CULL_FACE)

        if self.gl_api == "gl":
            if self.PROGRAM_POINT_SIZE in self._flags:
                gl.glEnable(self.PROGRAM_POINT_SIZE)
            else:
                gl.glDisable(self.PROGRAM_POINT_SIZE)

    @contextmanager
    def enabled(self, *flags):
        """
        Temporarily change enabled flags.

        Flags that was enabled initially will stay enabled.
        Only new enabled flags will be reversed when exiting
        the context.

        Example::

            with ctx.enabled(ctx.BLEND, ctx.CULL_FACE):
                # Render something
        """
        flags = set(flags)
        new_flags = flags - self._flags

        self.enable(*flags)
        try:
            yield
        finally:
            self.disable(*new_flags)

    @contextmanager
    def enabled_only(self, *flags):
        """
        Temporarily change enabled flags.

        Only the supplied flags with be enabled in in the context. When exiting
        the context the old flags will be restored.

        Example::

            with ctx.enabled_only(ctx.BLEND, ctx.CULL_FACE):
                # Render something
        """
        old_flags = self._flags
        self.enable_only(*flags)
        try:
            yield
        finally:
            self.enable_only(*old_flags)

    def disable(self, *args):
        """
        Disable one or more context flags::

            # Single flag
            ctx.disable(ctx.BLEND)
            # Multiple flags
            ctx.disable(ctx.DEPTH_TEST, ctx.CULL_FACE)
        """
        self._flags -= set(args)

        for flag in args:
            gl.glDisable(flag)

    def is_enabled(self, flag) -> bool:
        """
        Check if a context flag is enabled.

        .. Warning::

            This only tracks states set through this context instance.
            It does not query the actual OpenGL state. If you change context
            flags by calling ``glEnable`` or ``glDisable`` directly you
            are on your own.
        """
        return flag in self._flags

    @property
    def viewport(self) -> Tuple[int, int, int, int]:
        """
        Get or set the viewport for the currently active framebuffer.
        The viewport simply describes what pixels of the screen
        OpenGL should render to. Normally it would be the size of
        the window's framebuffer::

            # 4:3 screen
            ctx.viewport = 0, 0, 800, 600
            # 1080p
            ctx.viewport = 0, 0, 1920, 1080
            # Using the current framebuffer size
            ctx.viewport = 0, 0, *ctx.screen.size
        """
        return self.active_framebuffer.viewport

    @viewport.setter
    def viewport(self, value: Tuple[int, int, int, int]):
        self.active_framebuffer.viewport = value

    @property
    def scissor(self) -> Tuple[int, int, int, int] | None:
        """
        Get or set the scissor box for the active framebuffer.
        This is a shortcut for :py:meth:`~arcade.gl.Framebuffer.scissor`.

        By default the scissor box is disabled and has no effect
        and will have an initial value of ``None``. The scissor
        box is enabled when setting a value and disabled when
        set to ``None``.

        Example::

            # Set and enable scissor box only drawing
            # in a 100 x 100 pixel lower left area
            ctx.scissor = 0, 0, 100, 100
            # Disable scissoring
            ctx.scissor = None
        """
        return self.fbo.scissor

    @scissor.setter
    def scissor(self, value):
        self.fbo.scissor = value

    @property
    def blend_func(self) -> Tuple[int, int] | Tuple[int, int, int, int]:
        """
        Get or set the blend function.
        This is tuple specifying how the color and
        alpha blending factors are computed for the source
        and destination pixel.

        When using a two component tuple you specify the
        blend function for the source and the destination.

        When using a four component tuple you specify the
        blend function for the source color, source alpha
        destination color and destination alpha. (separate blend
        functions for color and alpha)

        Supported blend functions are::

            ZERO
            ONE
            SRC_COLOR
            ONE_MINUS_SRC_COLOR
            DST_COLOR
            ONE_MINUS_DST_COLOR
            SRC_ALPHA
            ONE_MINUS_SRC_ALPHA
            DST_ALPHA
            ONE_MINUS_DST_ALPHA

            # Shortcuts
            DEFAULT_BLENDING     # (SRC_ALPHA, ONE_MINUS_SRC_ALPHA)
            ADDITIVE_BLENDING    # (ONE, ONE)
            PREMULTIPLIED_ALPHA  # (SRC_ALPHA, ONE)

        These enums can be accessed in the ``arcade.gl``
        module or simply as attributes of the context object.
        The raw enums from ``pyglet.gl`` can also be used.

        Example::

            # Using constants from the context object
            ctx.blend_func = ctx.ONE, ctx.ONE
            # from the gl module
            from arcade import gl
            ctx.blend_func = gl.ONE, gl.ONE
        """
        return self._blend_func

    @blend_func.setter
    def blend_func(self, value: Tuple[int, int] | Tuple[int, int, int, int]):
        self._blend_func = value
        if len(value) == 2:
            gl.glBlendFunc(*value)
        elif len(value) == 4:
            gl.glBlendFuncSeparate(*value)
        else:
            ValueError("blend_func takes a tuple of 2 or 4 values")

    # def blend_equation(self)
    # Default is FUNC_ADD

    @property
    def front_face(self) -> str:
        """
        Configure front face winding order of triangles.

        By default the counter-clockwise winding side is the front face.
        This can be set set to clockwise or counter-clockwise::

            ctx.front_face = "cw"
            ctx.front_face = "ccw"
        """
        value = c_int()
        gl.glGetIntegerv(gl.GL_FRONT_FACE, value)
        return "cw" if value.value == gl.GL_CW else "ccw"

    @front_face.setter
    def front_face(self, value: str):
        if value not in ["cw", "ccw"]:
            raise ValueError("front_face must be 'cw' or 'ccw'")
        gl.glFrontFace(gl.GL_CW if value == "cw" else gl.GL_CCW)

    @property
    def cull_face(self) -> str:
        """
        The face side to cull when face culling is enabled.

        By default the back face is culled. This can be set to
        front, back or front_and_back::

            ctx.cull_face = "front"
            ctx.cull_face = "back"
            ctx.cull_face = "front_and_back"
        """
        value = c_int()
        gl.glGetIntegerv(gl.GL_CULL_FACE_MODE, value)
        return self._cull_face_options_reverse[value.value]

    @cull_face.setter
    def cull_face(self, value):
        if value not in self._cull_face_options:
            raise ValueError("cull_face must be", list(self._cull_face_options.keys()))

        gl.glCullFace(self._cull_face_options[value])

    @property
    def wireframe(self) -> bool:
        """
        Get or set the wireframe mode.

        When enabled all primitives will be rendered as lines
        by changing the polygon mode.
        """
        return self._wireframe

    @wireframe.setter
    def wireframe(self, value: bool):
        self._wireframe = value
        if value:
            gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_LINE)
        else:
            gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_FILL)

    @property
    def patch_vertices(self) -> int:
        """
        Get or set number of vertices that will be used to make up a single patch primitive.

        Patch primitives are consumed by the tessellation control shader (if present)
        and subsequently used for tessellation.
        """
        value = c_int()
        gl.glGetIntegerv(gl.GL_PATCH_VERTICES, value)
        return value.value

    @patch_vertices.setter
    def patch_vertices(self, value: int):
        if not isinstance(value, int):
            raise TypeError("patch_vertices must be an integer")

        gl.glPatchParameteri(gl.GL_PATCH_VERTICES, value)

    @property
    def point_size(self) -> float:
        """
        Set or get the point size. Default is `1.0`.

        Point size changes the pixel size of rendered points. The min and max values
        are limited by :py:attr:`~arcade.gl.context.Limits.POINT_SIZE_RANGE`.
        This value usually at least ``(1, 100)``, but this depends on the drivers/vendors.

        If variable point size is needed you can enable
        :py:attr:`~arcade.gl.Context.PROGRAM_POINT_SIZE` and write to ``gl_PointSize``
        in the vertex or geometry shader.

        .. Note::

            Using a geometry shader to create triangle strips from points is often a safer
            way to render large points since you don't have have any size restrictions
            and it offers more flexibility.
        """
        return self._point_size

    @point_size.setter
    def point_size(self, value: float):
        if self.gl_api == "gl":
            gl.glPointSize(self._point_size)
        self._point_size = value

    @property
    def primitive_restart_index(self) -> int:
        """
        Get or set the primitive restart index. Default is ``-1``.

        The primitive restart index can be used in index buffers
        to restart a primitive. This is for example useful when you
        use triangle strips or line strips and want to start on
        a new strip in the same buffer / draw call.
        """
        return self._primitive_restart_index

    @primitive_restart_index.setter
    def primitive_restart_index(self, value: int):
        self._primitive_restart_index = value
        if self.gl_api == "gl":
            gl.glPrimitiveRestartIndex(value)

    def finish(self) -> None:
        """
        Wait until all OpenGL rendering commands are completed.

        This function will actually stall until all work is done
        and may have severe performance implications.
        """
        gl.glFinish()

    def flush(self) -> None:
        """
        Flush the OpenGL command buffer.

        This will send all queued commands to the GPU but will not wait
        until they are completed. This is useful when you want to
        ensure that all commands are sent to the GPU before doing
        something else.
        """
        gl.glFlush()

    # Various utility methods

    def copy_framebuffer(
        self,
        src: Framebuffer,
        dst: Framebuffer,
        src_attachment_index: int = 0,
        depth: bool = True,
    ):
        """
        Copies/blits a framebuffer to another one.
        We can select one color attachment to copy plus
        an optional depth attachment.

        This operation has many restrictions to ensure it works across
        different platforms and drivers:

        * The source and destination framebuffer must be the same size
        * The formats of the attachments must be the same
        * Only the source framebuffer can be multisampled
        * Framebuffers cannot have integer attachments

        Args:
            src:
                The framebuffer to copy from
            dst:
                The framebuffer we copy to
            src_attachment_index:
                The color attachment to copy from
            depth:
                Also copy depth attachment if present
        """
        # Set source and dest framebuffer
        gl.glBindFramebuffer(gl.GL_READ_FRAMEBUFFER, src._glo)
        gl.glBindFramebuffer(gl.GL_DRAW_FRAMEBUFFER, dst._glo)

        # TODO: We can support blitting multiple layers here
        gl.glReadBuffer(gl.GL_COLOR_ATTACHMENT0 + src_attachment_index)
        if dst.is_default:
            gl.glDrawBuffer(gl.GL_BACK)
        else:
            gl.glDrawBuffer(gl.GL_COLOR_ATTACHMENT0)

        # gl.glBindFramebuffer(gl.GL_FRAMEBUFFER, src._glo)
        gl.glBlitFramebuffer(
            0,
            0,
            src.width,
            src.height,  # Make source and dest size the same
            0,
            0,
            src.width,
            src.height,
            gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT,
            gl.GL_NEAREST,
        )

        # Reset states. We can also apply previous states here
        gl.glReadBuffer(gl.GL_COLOR_ATTACHMENT0)

    # --- Resource methods ---

    def buffer(
        self, *, data: BufferProtocol | None = None, reserve: int = 0, usage: str = "static"
    ) -> Buffer:
        """
        Create an OpenGL Buffer object. The buffer will contain all zero-bytes if
        no data is supplied.

        Examples::

            # Create 1024 byte buffer
            ctx.buffer(reserve=1024)
            # Create a buffer with 1000 float values using python's array.array
            from array import array
            ctx.buffer(data=array('f', [i for in in range(1000)])
            # Create a buffer with 1000 random 32 bit floats using numpy
            self.ctx.buffer(data=np.random.random(1000).astype("f4"))

        The ``data`` parameter can be anything that implements the
        `Buffer Protocol <https://docs.python.org/3/c-api/buffer.html>`_.

        This includes ``bytes``, ``bytearray``, ``array.array``, and
        more. You may need to use typing workarounds for non-builtin
        types. See :ref:`prog-guide-gl-buffer-protocol-typing` for more
        information.

        The ``usage`` parameter enables the GL implementation to make more intelligent
        decisions that may impact buffer object performance. It does not add any restrictions.
        If in doubt, skip this parameter and revisit when optimizing. The result
        are likely to be different between vendors/drivers or may not have any effect.
        Always use the default static usage for buffers that don't change.

        The available values mean the following::

            stream
                The data contents will be modified once and used at most a few times.
            static
                The data contents will be modified once and used many times.
            dynamic
                The data contents will be modified repeatedly and used many times.

        Args:
            data:
                The buffer data. This can be a ``bytes`` instance or any
                any other object supporting the buffer protocol.
            reserve:
                The number of bytes to reserve
            usage:
                Buffer usage. 'static', 'dynamic' or 'stream'
        """
        return Buffer(self, data, reserve=reserve, usage=usage)

    def framebuffer(
        self,
        *,
        color_attachments: Texture2D | List[Texture2D] | None = None,
        depth_attachment: Texture2D | None = None,
    ) -> Framebuffer:
        """Create a Framebuffer.

        Args:
            color_attachments:
                List of textures we want to render into
            depth_attachment:
                Depth texture
        """
        return Framebuffer(
            self, color_attachments=color_attachments or [], depth_attachment=depth_attachment
        )

    def texture(
        self,
        size: Tuple[int, int],
        *,
        components: int = 4,
        dtype: str = "f1",
        data: BufferProtocol | None = None,
        wrap_x: PyGLenum | None = None,
        wrap_y: PyGLenum | None = None,
        filter: Tuple[PyGLenum, PyGLenum] | None = None,
        samples: int = 0,
        immutable: bool = False,
        internal_format: PyGLenum | None = None,
        compressed: bool = False,
        compressed_data: bool = False,
    ) -> Texture2D:
        """
        Create a 2D Texture.

        Example::

            # Create a 1024 x 1024 RGBA texture
            image = PIL.Image.open("my_texture.png")
            ctx.texture(size=(1024, 1024), components=4, data=image.tobytes())

            # Create and compress a texture. The compression format is set by the internal_format
            image = PIL.Image.open("my_texture.png")
            ctx.texture(
                size=(1024, 1024),
                components=4,
                compressed=True,
                internal_format=gl.GL_COMPRESSED_RGBA_S3TC_DXT1_EXT,
                data=image.tobytes(),
            )

            # Create a compressed texture from raw compressed data. This is an extremely
            # fast way to load a large number of textures.
            image_bytes = "<raw compressed data from some source>"
            ctx.texture(
                size=(1024, 1024),
                components=4,
                internal_format=gl.GL_COMPRESSED_RGBA_S3TC_DXT1_EXT,
                compressed_data=True,
                data=image_bytes,
            )

        Wrap modes: ``GL_REPEAT``, ``GL_MIRRORED_REPEAT``, ``GL_CLAMP_TO_EDGE``,
        ``GL_CLAMP_TO_BORDER``

        Minifying filters: ``GL_NEAREST``, ``GL_LINEAR``, ``GL_NEAREST_MIPMAP_NEAREST``,
        ``GL_LINEAR_MIPMAP_NEAREST`` ``GL_NEAREST_MIPMAP_LINEAR``, ``GL_LINEAR_MIPMAP_LINEAR``

        Magnifying filters: ``GL_NEAREST``, ``GL_LINEAR``

        Args:
            size:
                The size of the texture
            components:
                Number of components (1: R, 2: RG, 3: RGB, 4: RGBA)
            dtype:
                The data type of each component: f1, f2, f4 / i1, i2, i4 / u1, u2, u4
            data:
                The texture data. Can be ``bytes``
                or any object supporting the buffer protocol.
            wrap_x:
                How the texture wraps in x direction
            wrap_y:
                How the texture wraps in y direction
            filter:
                Minification and magnification filter
            samples:
                Creates a multisampled texture for values > 0
            immutable:
                Make the storage (not the contents) immutable. This can sometimes be
                required when using textures with compute shaders.
            internal_format:
                The internal format of the texture. This can be used to
                enable sRGB or texture compression.
            compressed:
                Set to True if you want the texture to be compressed.
                This assumes you have set a internal_format to a compressed format.
            compressed_data:
                Set to True if you are passing in raw compressed pixel data.
                This implies ``compressed=True``.
        """
        compressed = compressed or compressed_data

        return Texture2D(
            self,
            size,
            components=components,
            data=data,
            dtype=dtype,
            wrap_x=wrap_x,
            wrap_y=wrap_y,
            filter=filter,
            samples=samples,
            immutable=immutable,
            internal_format=internal_format,
            compressed=compressed,
            compressed_data=compressed_data,
        )

    def texture_array(
        self,
        size: Tuple[int, int, int],
        *,
        components: int = 4,
        dtype: str = "f1",
        data: BufferProtocol | None = None,
        wrap_x: PyGLenum | None = None,
        wrap_y: PyGLenum | None = None,
        filter: Tuple[PyGLenum, PyGLenum] | None = None,
    ) -> TextureArray:
        """
        Create a 2D Texture Array.

        This is a 2D texture with multiple layers. This is useful for
        storing multiple textures in a single texture object. This can
        be used for texture atlases or storing multiple frames of an
        animation in a single texture or equally sized tile textures.

        Note that ``size`` is a 3-tuple where the last value is the number  of layers.

        See :py:meth:`~arcade.gl.Context.texture` for arguments.
        """
        return TextureArray(
            self,
            size,
            components=components,
            dtype=dtype,
            data=data,
            wrap_x=wrap_x,
            wrap_y=wrap_y,
            filter=filter,
        )

    def depth_texture(
        self, size: Tuple[int, int], *, data: BufferProtocol | None = None
    ) -> Texture2D:
        """
        Create a 2D depth texture. Can be used as a depth attachment
        in a :py:class:`~arcade.gl.Framebuffer`.

        Args:
            size:
                The size of the texture
            data:
                The texture data. Can be``bytes`` or any object
                supporting the buffer protocol.
        """
        return Texture2D(self, size, data=data, depth=True)

    def sampler(self, texture: Texture2D) -> Sampler:
        """
        Create a sampler object for a texture.

        Args:
            texture:
                The texture to create a sampler for
        """
        return Sampler(self, texture)

    def geometry(
        self,
        content: Sequence[BufferDescription] | None = None,
        index_buffer: Buffer | None = None,
        mode: int | None = None,
        index_element_size: int = 4,
    ):
        """
        Create a Geometry instance. This is Arcade's version of a vertex array adding
        a lot of convenience for the user. Geometry objects are fairly light. They are
        mainly responsible for automatically map buffer inputs to your shader(s)
        and provide various methods for rendering or processing this geometry,

        The same geometry can be rendered with different
        programs as long as your shader is using one or more of the input attribute.
        This means geometry with positions and colors can be rendered with a program
        only using the positions. We will automatically map what is necessary and
        cache these mappings internally for performance.

        In short, the geometry object is a light object that describes what buffers
        contains and automatically negotiate with shaders/programs. This is a very
        complex field in OpenGL so the Geometry object provides substantial time
        savings and greatly reduces the complexity of your code.

        Geometry also provide rendering methods supporting the following:

        * Rendering geometry with and without index buffer
        * Rendering your geometry using instancing. Per instance buffers can be provided
          or the current instance can be looked up using ``gl_InstanceID`` in shaders.
        * Running transform feedback shaders that writes to buffers instead the screen.
          This can write to one or multiple buffer.
        * Render your geometry with indirect rendering. This means packing
          multiple meshes into the same buffer(s) and batch drawing them.

        Examples::

            # Single buffer geometry with a vec2 vertex position attribute
            ctx.geometry([BufferDescription(buffer, '2f', ["in_vert"])], mode=ctx.TRIANGLES)

            # Single interleaved buffer with two attributes. A vec2 position and vec2 velocity
            ctx.geometry([
                    BufferDescription(buffer, '2f 2f', ["in_vert", "in_velocity"])
                ],
                mode=ctx.POINTS,
            )

            # Geometry with index buffer
            ctx.geometry(
                [BufferDescription(buffer, '2f', ["in_vert"])],
                index_buffer=ibo,
                mode=ctx.TRIANGLES,
            )

            # Separate buffers
            ctx.geometry([
                    BufferDescription(buffer_pos, '2f', ["in_vert"])
                    BufferDescription(buffer_vel, '2f', ["in_velocity"])
                ],
                mode=ctx.POINTS,
            )

            # Providing per-instance data for instancing
            ctx.geometry([
                    BufferDescription(buffer_pos, '2f', ["in_vert"])
                    BufferDescription(buffer_instance_pos, '2f', ["in_offset"], instanced=True)
                ],
                mode=ctx.POINTS,
            )

        Args:
            content:
                List of :py:class:`~arcade.gl.BufferDescription`
            index_buffer:
                Index/element buffer
            mode:
                The default draw mode
            mode:
                The default draw mode
            index_element_size:
                Byte size of a single index/element in the index buffer.
                In other words, the index buffer can be 1, 2 or 4 byte integers.
                Can be 1, 2 or 4 (8, 16 or 32 bit unsigned integer)
        """
        return Geometry(
            self,
            content,
            index_buffer=index_buffer,
            mode=mode,
            index_element_size=index_element_size,
        )

    def program(
        self,
        *,
        vertex_shader: str,
        fragment_shader: str | None = None,
        geometry_shader: str | None = None,
        tess_control_shader: str | None = None,
        tess_evaluation_shader: str | None = None,
        common: List[str] | None = None,
        defines: Dict[str, str] | None = None,
        varyings: Sequence[str] | None = None,
        varyings_capture_mode: str = "interleaved",
    ) -> Program:
        """
        Create a :py:class:`~arcade.gl.Program` given shader sources
        and other settings.

        Args:
            vertex_shader:
                vertex shader source
            fragment_shader:
                fragment shader source
            geometry_shader:
                geometry shader source
            tess_control_shader:
                tessellation control shader source
            tess_evaluation_shader:
                tessellation evaluation shader source
            common:
                Common shader sources injected into all shaders
            defines:
                Substitute #defines values in the source
            varyings:
                The name of the out attributes in a transform shader.
                This is normally not necessary since we auto detect them,
                but some more complex out structures we can't detect.
            varyings_capture_mode:
                The capture mode for transforms.

                - ``"interleaved"`` means all out attribute will be written to a single buffer.
                - ``"separate"`` means each out attribute will be written separate buffers.

                Based on these settings the ``transform()`` method will accept a single
                buffer or a list of buffer.
        """
        source_vs = ShaderSource(self, vertex_shader, common, gl.GL_VERTEX_SHADER)
        source_fs = (
            ShaderSource(self, fragment_shader, common, gl.GL_FRAGMENT_SHADER)
            if fragment_shader
            else None
        )
        source_geo = (
            ShaderSource(self, geometry_shader, common, gl.GL_GEOMETRY_SHADER)
            if geometry_shader
            else None
        )
        source_tc = (
            ShaderSource(self, tess_control_shader, common, gl.GL_TESS_CONTROL_SHADER)
            if tess_control_shader
            else None
        )
        source_te = (
            ShaderSource(self, tess_evaluation_shader, common, gl.GL_TESS_EVALUATION_SHADER)
            if tess_evaluation_shader
            else None
        )

        # If we don't have a fragment shader we are doing transform feedback.
        # When a geometry shader is present the out attributes will be located there
        out_attributes = list(varyings) if varyings is not None else []  # type: List[str]
        if not source_fs and not out_attributes:
            if source_geo:
                out_attributes = source_geo.out_attributes
            else:
                out_attributes = source_vs.out_attributes

        return Program(
            self,
            vertex_shader=source_vs.get_source(defines=defines),
            fragment_shader=source_fs.get_source(defines=defines) if source_fs else None,
            geometry_shader=source_geo.get_source(defines=defines) if source_geo else None,
            tess_control_shader=source_tc.get_source(defines=defines) if source_tc else None,
            tess_evaluation_shader=source_te.get_source(defines=defines) if source_te else None,
            varyings=out_attributes,
            varyings_capture_mode=varyings_capture_mode,
        )

    def query(self, *, samples=True, time=True, primitives=True) -> Query:
        """
        Create a query object for measuring rendering calls in opengl.

        Args:
            samples: Collect written samples
            time: Measure rendering duration
            primitives: Collect the number of primitives emitted
        """
        return Query(self, samples=samples, time=time, primitives=primitives)

    def compute_shader(self, *, source: str, common: Iterable[str] = ()) -> ComputeShader:
        """
        Create a compute shader.

        Args:
            source:
                The glsl source
            common:
                Common / library source injected into compute shader
        """
        src = ShaderSource(self, source, common, gl.GL_COMPUTE_SHADER)
        return ComputeShader(self, src.get_source())


class ContextStats:
    """
    Runtime allocation statistics of OpenGL objects.
    """

    def __init__(self, warn_threshold=100):
        self.warn_threshold = warn_threshold

        self.texture = (0, 0)
        """Textures (created, freed)"""

        self.framebuffer = (0, 0)
        """Framebuffers (created, freed)"""

        self.buffer = (0, 0)
        """Buffers (created, freed)"""

        self.program = (0, 0)
        """Programs (created, freed)"""

        self.vertex_array = (0, 0)
        """Vertex Arrays (created, freed)"""

        self.geometry = (0, 0)
        """Geometry (created, freed)"""

        self.compute_shader = (0, 0)
        """Compute Shaders (created, freed)"""

        self.query = (0, 0)
        """Queries (created, freed)"""

    def incr(self, key: str) -> None:
        """
        Increments a counter.

        Args:
            key: The attribute name / counter to increment.
        """
        created, freed = getattr(self, key)
        setattr(self, key, (created + 1, freed))
        if created % self.warn_threshold == 0 and created > 0:
            LOG.debug(
                "%s allocations passed threshold (%s) [created = %s] [freed = %s] [active = %s]",
                key,
                self.warn_threshold,
                created,
                freed,
                created - freed,
            )

    def decr(self, key):
        """
        Decrement a counter.

        Args:
            key: The attribute name / counter to decrement.
        """
        created, freed = getattr(self, key)
        setattr(self, key, (created, freed + 1))


class GLInfo:
    """OpenGL info and capabilities"""

    def __init__(self, ctx):
        self._ctx = ctx

        self.MINOR_VERSION = self.get(gl.GL_MINOR_VERSION)
        """Minor version number of the OpenGL API supported by the current context"""

        self.MAJOR_VERSION = self.get(gl.GL_MAJOR_VERSION)
        """Major version number of the OpenGL API supported by the current context."""

        self.VENDOR = self.get_str(gl.GL_VENDOR)
        """The vendor string. For example 'NVIDIA Corporation'"""

        self.RENDERER = self.get_str(gl.GL_RENDERER)
        """The renderer things. For example "NVIDIA GeForce RTX 2080 SUPER/PCIe/SSE2"""

        self.SAMPLE_BUFFERS = self.get(gl.GL_SAMPLE_BUFFERS)
        """Value indicating the number of sample buffers associated with the framebuffer"""

        self.SUBPIXEL_BITS = self.get(gl.GL_SUBPIXEL_BITS)
        """
        An estimate of the number of bits of subpixel resolution
        that are used to position rasterized geometry in window coordinates
        """

        self.UNIFORM_BUFFER_OFFSET_ALIGNMENT = self.get(gl.GL_UNIFORM_BUFFER_OFFSET_ALIGNMENT)
        """Minimum required alignment for uniform buffer sizes and offset"""

        self.MAX_ARRAY_TEXTURE_LAYERS = self.get(gl.GL_MAX_ARRAY_TEXTURE_LAYERS)
        """
        Value indicates the maximum number of layers allowed in an array texture,
        and must be at least 256
        """

        self.MAX_3D_TEXTURE_SIZE = self.get(gl.GL_MAX_3D_TEXTURE_SIZE)
        """
        A rough estimate of the largest 3D texture that the GL can handle.
        The value must be at least 64
        """

        self.MAX_COLOR_ATTACHMENTS = self.get(gl.GL_MAX_COLOR_ATTACHMENTS)
        """Maximum number of color attachments in a framebuffer"""

        self.MAX_COLOR_TEXTURE_SAMPLES = self.get(gl.GL_MAX_COLOR_TEXTURE_SAMPLES)
        """Maximum number of samples in a color multisample texture"""

        self.MAX_COMBINED_FRAGMENT_UNIFORM_COMPONENTS = self.get(
            gl.GL_MAX_COMBINED_FRAGMENT_UNIFORM_COMPONENTS
        )
        """the number of words for fragment shader uniform variables in all uniform blocks"""

        self.MAX_COMBINED_GEOMETRY_UNIFORM_COMPONENTS = self.get(
            gl.GL_MAX_COMBINED_GEOMETRY_UNIFORM_COMPONENTS
        )
        """Number of words for geometry shader uniform variables in all uniform blocks"""

        self.MAX_COMBINED_TEXTURE_IMAGE_UNITS = self.get(gl.GL_MAX_COMBINED_TEXTURE_IMAGE_UNITS)
        """
        Maximum supported texture image units that can be used to access texture
        maps from the vertex shader
        """

        self.MAX_COMBINED_UNIFORM_BLOCKS = self.get(gl.GL_MAX_COMBINED_UNIFORM_BLOCKS)
        """Maximum number of uniform blocks per program"""

        self.MAX_COMBINED_VERTEX_UNIFORM_COMPONENTS = self.get(
            gl.GL_MAX_COMBINED_VERTEX_UNIFORM_COMPONENTS
        )
        """Number of words for vertex shader uniform variables in all uniform blocks"""

        self.MAX_CUBE_MAP_TEXTURE_SIZE = self.get(gl.GL_MAX_CUBE_MAP_TEXTURE_SIZE)
        """A rough estimate of the largest cube-map texture that the GL can handle"""

        self.MAX_DEPTH_TEXTURE_SAMPLES = self.get(gl.GL_MAX_DEPTH_TEXTURE_SAMPLES)
        """Maximum number of samples in a multisample depth or depth-stencil texture"""

        self.MAX_DRAW_BUFFERS = self.get(gl.GL_MAX_DRAW_BUFFERS)
        """Maximum number of simultaneous outputs that may be written in a fragment shader"""

        self.MAX_ELEMENTS_INDICES = self.get(gl.GL_MAX_ELEMENTS_INDICES)
        """Recommended maximum number of vertex array indices"""

        self.MAX_ELEMENTS_VERTICES = self.get(gl.GL_MAX_ELEMENTS_VERTICES)
        """Recommended maximum number of vertex array vertices"""

        self.MAX_FRAGMENT_INPUT_COMPONENTS = self.get(gl.GL_MAX_FRAGMENT_INPUT_COMPONENTS)
        """Maximum number of components of the inputs read by the fragment shader"""

        self.MAX_FRAGMENT_UNIFORM_COMPONENTS = self.get(gl.GL_MAX_FRAGMENT_UNIFORM_COMPONENTS)
        """
        Maximum number of individual floating-point, integer, or boolean values that can be
        held in uniform variable storage for a fragment shader
        """

        self.MAX_FRAGMENT_UNIFORM_VECTORS = self.get(gl.GL_MAX_FRAGMENT_UNIFORM_VECTORS)
        """
        Maximum number of individual 4-vectors of floating-point, integer,
        or boolean values that can be held in uniform variable storage for a fragment shader
        """

        self.MAX_FRAGMENT_UNIFORM_BLOCKS = self.get(gl.GL_MAX_FRAGMENT_UNIFORM_BLOCKS)
        """Maximum number of uniform blocks per fragment shader."""

        self.MAX_GEOMETRY_INPUT_COMPONENTS = self.get(gl.GL_MAX_GEOMETRY_INPUT_COMPONENTS)
        """Maximum number of components of inputs read by a geometry shader"""

        self.MAX_GEOMETRY_OUTPUT_COMPONENTS = self.get(gl.GL_MAX_GEOMETRY_OUTPUT_COMPONENTS)
        """Maximum number of components of outputs written by a geometry shader"""

        self.MAX_GEOMETRY_TEXTURE_IMAGE_UNITS = self.get(gl.GL_MAX_GEOMETRY_TEXTURE_IMAGE_UNITS)
        """
        Maximum supported texture image units that can be used to access texture
        maps from the geometry shader
        """

        self.MAX_GEOMETRY_UNIFORM_BLOCKS = self.get(gl.GL_MAX_GEOMETRY_UNIFORM_BLOCKS)
        """Maximum number of uniform blocks per geometry shader"""

        self.MAX_GEOMETRY_UNIFORM_COMPONENTS = self.get(gl.GL_MAX_GEOMETRY_UNIFORM_COMPONENTS)
        """
        Maximum number of individual floating-point, integer, or boolean values that can
        be held in uniform variable storage for a geometry shader
        """

        self.MAX_INTEGER_SAMPLES = self.get(gl.GL_MAX_INTEGER_SAMPLES)
        """Maximum number of samples supported in integer format multisample buffers"""

        self.MAX_SAMPLES = self.get(gl.GL_MAX_SAMPLES)
        """Maximum samples for a framebuffer"""

        self.MAX_RENDERBUFFER_SIZE = self.get(gl.GL_MAX_RENDERBUFFER_SIZE)
        """Maximum supported size for renderbuffers"""

        self.MAX_SAMPLE_MASK_WORDS = self.get(gl.GL_MAX_SAMPLE_MASK_WORDS)
        """Maximum number of sample mask words"""

        self.MAX_UNIFORM_BUFFER_BINDINGS = self.get(gl.GL_MAX_UNIFORM_BUFFER_BINDINGS)
        """Maximum number of uniform buffer binding points on the context"""

        self.MAX_UNIFORM_BUFFER_BINDINGS = self.get(gl.GL_MAX_UNIFORM_BUFFER_BINDINGS)
        """Maximum number of uniform buffer binding points on the context"""

        self.MAX_TEXTURE_SIZE = self.get(gl.GL_MAX_TEXTURE_SIZE)
        """The value gives a rough estimate of the largest texture that the GL can handle"""

        self.MAX_UNIFORM_BUFFER_BINDINGS = self.get(gl.GL_MAX_UNIFORM_BUFFER_BINDINGS)
        """Maximum number of uniform buffer binding points on the context"""

        self.MAX_UNIFORM_BLOCK_SIZE = self.get(gl.GL_MAX_UNIFORM_BLOCK_SIZE)
        """Maximum size in basic machine units of a uniform block"""

        self.MAX_VARYING_VECTORS = self.get(gl.GL_MAX_VARYING_VECTORS)
        """The number 4-vectors for varying variables"""

        self.MAX_VERTEX_ATTRIBS = self.get(gl.GL_MAX_VERTEX_ATTRIBS)
        """Maximum number of 4-component generic vertex attributes accessible to a vertex shader."""

        self.MAX_VERTEX_TEXTURE_IMAGE_UNITS = self.get(gl.GL_MAX_VERTEX_TEXTURE_IMAGE_UNITS)
        """
        Maximum supported texture image units that can be used to access texture
        maps from the vertex shader.
        """

        self.MAX_VERTEX_UNIFORM_COMPONENTS = self.get(gl.GL_MAX_VERTEX_UNIFORM_COMPONENTS)
        """
        Maximum number of individual floating-point, integer, or boolean values that
        can be held in uniform variable storage for a vertex shader
        """

        self.MAX_VERTEX_UNIFORM_VECTORS = self.get(gl.GL_MAX_VERTEX_UNIFORM_VECTORS)
        """
        Maximum number of 4-vectors that may be held in uniform variable storage
        for the vertex shader
        """

        self.MAX_VERTEX_OUTPUT_COMPONENTS = self.get(gl.GL_MAX_VERTEX_OUTPUT_COMPONENTS)
        """Maximum number of components of output written by a vertex shader"""

        self.MAX_VERTEX_UNIFORM_BLOCKS = self.get(gl.GL_MAX_VERTEX_UNIFORM_BLOCKS)
        """Maximum number of uniform blocks per vertex shader."""

        # self.MAX_VERTEX_ATTRIB_RELATIVE_OFFSET = self.get(
        #     gl.GL_MAX_VERTEX_ATTRIB_RELATIVE_OFFSET
        # )
        # self.MAX_VERTEX_ATTRIB_BINDINGS = self.get(gl.GL_MAX_VERTEX_ATTRIB_BINDINGS)

        self.MAX_TEXTURE_IMAGE_UNITS = self.get(gl.GL_MAX_TEXTURE_IMAGE_UNITS)
        """Number of texture units"""

        self.MAX_TEXTURE_MAX_ANISOTROPY = self.get_float(gl.GL_MAX_TEXTURE_MAX_ANISOTROPY, 1.0)
        """The highest supported anisotropy value. Usually 8.0 or 16.0."""

        self.MAX_VIEWPORT_DIMS: Tuple[int, int] = self.get_int_tuple(gl.GL_MAX_VIEWPORT_DIMS, 2)
        """
        The maximum support window or framebuffer viewport.
        This is usually the same as the maximum texture size
        """

        self.MAX_TRANSFORM_FEEDBACK_SEPARATE_ATTRIBS = self.get(
            gl.GL_MAX_TRANSFORM_FEEDBACK_SEPARATE_ATTRIBS
        )
        """
        How many buffers we can have as output when doing a transform(feedback).
        This is usually 4.
        """

        self.POINT_SIZE_RANGE = self.get_int_tuple(gl.GL_POINT_SIZE_RANGE, 2)
        """The minimum and maximum point size"""

        err = self._ctx.error
        if err:
            from warnings import warn

            warn("Error happened while querying of limits. Moving on ..")

    @overload
    def get_int_tuple(self, enum: GLenumLike, length: Literal[2]) -> Tuple[int, int]: ...

    @overload
    def get_int_tuple(self, enum: GLenumLike, length: int) -> Tuple[int, ...]: ...

    def get_int_tuple(self, enum: GLenumLike, length: int):
        """
        Get an enum as an int tuple

        Args:
            enum: The enum to query
            length: The length of the tuple
        """
        try:
            values = (c_int * length)()
            gl.glGetIntegerv(enum, values)
            return tuple(values)
        except pyglet.gl.lib.GLException:
            return tuple([0] * length)

    def get(self, enum: GLenumLike, default=0) -> int:
        """
        Get an integer limit.

        Args:
            enum: The enum to query
            default: The default value if the query fails
        """
        try:
            value = c_int()
            gl.glGetIntegerv(enum, value)
            return value.value
        except pyglet.gl.lib.GLException:
            return default

    def get_float(self, enum: GLenumLike, default=0.0) -> float:
        """
        Get a float limit

        Args:
            enum: The enum to query
            default: The default value if the query fails
        """
        try:
            value = c_float()
            gl.glGetFloatv(enum, value)
            return value.value
        except pyglet.gl.lib.GLException:
            return default

    def get_str(self, enum: GLenumLike) -> str:
        """
        Get a string limit.

        Args:
            enum: The enum to query
        """
        try:
            return cast(gl.glGetString(enum), c_char_p).value.decode()  # type: ignore
        except pyglet.gl.lib.GLException:
            return "Unknown"



"""
Custom sprites with geometry shader

In some situations we could need some sprite-like geometry
that doesn't need the features of a sprite list. Initializing
structures for this kind of geometry is also less complex
and much faster than creation lots of sprite objects.

We are also free to design our own sprite layout with
any attribute we need. This data can be used to move
as much logic as possible into your shaders instead
of doing the logic in python.

In this example we're just using textured rectangles
to mimic sprites, but it could be circles or other
types of shapes with some imagination. We'll also be
limiting the number of attributes to make it simple.
Things like rotation, color, anchor etc can be added.

Hold and drag the mouse to scroll around.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.gl.custom_sprite
"""

from random import randint
from array import array

import arcade
from arcade.gl.types import BufferDescription


class GeoSprites(arcade.Window):

    def __init__(self):
        super().__init__(800, 600, "Custom Sprites", resizable=True)
        self.camera = arcade.Camera2D()
        self.program = self.ctx.program(
            vertex_shader="""
            #version 330

            // Input from our buffers
            in vec2 in_position;
            in vec2 in_size;

            // Outputs to our geometry shader
            out vec2 position;
            out vec2 size;

            void main() {
                // We just the information down unmodified
                position = in_position;
                size = in_size;
            }
            """,
            geometry_shader="""
            #version 330

            // Configure inputs and outputs for the geometry shader
            // We are taking single points form the vertex shader per invocation
            // and emitting 4 new vertices creating a quad/sprites
            layout (points) in;
            layout (triangle_strip, max_vertices = 4) out;

            // A uniform buffer that will automagically contain pyglet's projection matrix
            uniform WindowBlock {
                mat4 projection;
                mat4 view;
            } window;

            // Receive the outputs from the vertex shader.
            // Since geometry shader can take multiple values from a vertex
            // shader we need to define the inputs as arrays.
            // We're only getting one vertex at the time in this example,
            // but we make an unsized array leaving the rest up to the shader compiler.
            in vec2 position[];
            in vec2 size[];

            // Texture coordinate to fragment shader
            out vec2 uv;

            void main() {
                // Create some more convenient variables for the input
                vec2 center = position[0];
                vec2 hsize = size[0] / 2.0;

                // Emit a triangle strip of 4 vertices making a triangle.
                // The fragment shader will then fill these to triangles in the next stage.

                mat4 mvp = window.projection * window.view;

                // Upper left
                gl_Position = mvp * vec4(vec2(-hsize.x, hsize.y) + center, 0.0, 1.0);
                uv = vec2(0, 1);
                EmitVertex();

                // lower left
                gl_Position = mvp * vec4(vec2(-hsize.x, -hsize.y) + center, 0.0, 1.0);
                uv = vec2(0, 0);
                EmitVertex();

                // upper right
                gl_Position = mvp * vec4(vec2(hsize.x, hsize.y) + center, 0.0, 1.0);
                uv = vec2(1, 1);
                EmitVertex();

                // lower right
                gl_Position = mvp * vec4(vec2(hsize.x, -hsize.y) + center, 0.0, 1.0);
                uv = vec2(1, 0);
                EmitVertex();

                // We are done with this triangle strip now
                EndPrimitive();
            }
            """,
            fragment_shader="""
            #version 330

            // Fragment shader runs once for each pixel the triangles consists of.
            // The texture coordinate be an interpolated value based on
            // the location of the pixel in the triangle.

            // A sampler that can read from a texture channel
            uniform sampler2D sprite_texture;

            // Interpolated texture coordinates
            in vec2 uv;

            // The pixel that will be rendered to the screen
            out vec4 fragColor;

            void main() {
                // Look up a pixel in the texture and write to screen
                fragColor = texture(sprite_texture, uv);
            }
            """,
        )
        # Configure sampler to read from channel 0
        self.program["sprite_texture"] = 0
        # Load a image as an OpenGL texture
        self.texture = self.ctx.load_texture(":resources:images/tiles/boxCrate_double.png")

        self.num_sprites = 1000
        # Make an interleaved buffer with positions and sizes
        self.vertex_buffer = self.ctx.buffer(data=array("f", self.gen_sprites(self.num_sprites)))
        # Mage a geometry object describing the buffer contents for our shader
        self.geometry = self.ctx.geometry(
            content=[BufferDescription(self.vertex_buffer, "2f 2f", ["in_position", "in_size"])]
        )

    def on_resize(self, width: int, height: int):
        super().on_resize(width, height)
        self.camera.match_window(position=True)

    def on_draw(self):
        self.clear()
        self.camera.use()
        # Bind our sprite texture to channel 0
        self.texture.use(unit=0)
        # Render the sprite data with our shader
        self.geometry.render(self.program)

    def on_mouse_drag(self, x: float, y: float, dx: float, dy: float, buttons: int, modifiers: int):
        """Make it easier to explore the geometry by scrolling"""
        self.camera.position = self.camera.position[0] - dx, self.camera.position[1] - dy

    def gen_sprites(self, count: int):
        """Quickly generate some random sprite data"""
        for _ in range(count):
            # Position
            yield randint(0, self.width * 4)
            yield randint(0, self.height * 4)
            # Size (make squares for now, not rectangles)
            size = randint(20, 100)
            yield size
            yield size


GeoSprites().run()
